"""Implements a pipeline for importing reference data from a raster image."""
from typing import Optional

import fs
import numpy as np
from pydantic import Field

from eolearn.core import CreateEOPatchTask, EONode, EOWorkflow, FeatureType, OverwritePermission, SaveTask
from eolearn.features.feature_manipulation import SpatialResizeTask
from eolearn.features.utils import ResizeLib, ResizeMethod, ResizeParam
from eolearn.io import ImportFromTiffTask

from ..core.pipeline import Pipeline
from ..core.schemas import BaseSchema
from ..types import Feature, PatchList
from ..utils.filter import get_patches_with_missing_features
from ..utils.validators import ensure_storage_key_presence, optional_field_validator, parse_dtype


class ResizeSchema(BaseSchema):
    """How to resize the tiff data after adding it to EOPatches."""

    resize_type: ResizeParam = Field(
        description=(
            "Determines type of resizing process and how `width_param` and `height_param` are used. See"
            " `SpatialResizeTask` documentation for more info."
        )
    )
    width_param: float = Field(description="Parameter to be applied to the width in combination with the resize_type.")
    height_param: float = Field(
        description="Parameter to be applied to the height in combination with the resize_type."
    )
    method: ResizeMethod = ResizeMethod.LINEAR
    library: ResizeLib = ResizeLib.PIL


class ImportTiffPipeline(Pipeline):
    class Schema(Pipeline.Schema):
        output_folder_key: str = Field(description="The storage manager key of the output folder.")
        _ensure_output_folder_key = ensure_storage_key_presence("output_folder_key")

        tiff_folder_key: str = Field(
            "input_data",
            description="The storage manager key of the folder containing the tiff. Defaults to the input-data folder.",
        )
        _ensure_tiff_folder_key = ensure_storage_key_presence("tiff_folder_key")
        input_filename: str = Field(description="Name of tiff file to import.")
        output_feature: Feature = Field(description="Feature containing the imported tiff information.")
        no_data_value: float = Field(
            np.nan, description="Value assigned to undefined pixels, e.g. outside of given input image."
        )
        dtype: Optional[np.dtype] = Field(description="Custom dtype for the imported feature.")
        _parse_dtype = optional_field_validator("dtype", parse_dtype, pre=True)
        use_vsi: bool = Field(
            True,
            description="Whether to use the VSI for reading. Enabled by default as a remote filesystem is assumed.",
        )
        resize: Optional[ResizeSchema] = Field(
            description="Settings for SpatialResizeTask applied at the end. When omitted resizing is not performed."
        )

    config: Schema

    def filter_patch_list(self, patch_list: PatchList) -> PatchList:
        """EOPatches are filtered according to existence of new features."""
        filtered_patch_list = get_patches_with_missing_features(
            self.storage.filesystem,
            self.storage.get_folder(self.config.output_folder_key),
            patch_list,
            [self.config.output_feature, FeatureType.BBOX],
        )
        return filtered_patch_list

    def build_workflow(self) -> EOWorkflow:
        create_eopatch_node = EONode(CreateEOPatchTask())

        file_path = fs.path.join(self.storage.get_folder(self.config.tiff_folder_key), self.config.input_filename)
        import_task = ImportFromTiffTask(
            self.config.output_feature,
            file_path,
            no_data_value=self.config.no_data_value,
            filesystem=self.storage.filesystem,
            image_dtype=self.config.dtype,
            use_vsi=self.config.use_vsi,
        )
        import_node = EONode(import_task, inputs=[create_eopatch_node])

        resize_node = None
        if self.config.resize:
            width_param, height_param = self.config.resize.width_param, self.config.resize.height_param
            if self.config.resize.resize_type is ResizeParam.NEW_SIZE:
                # pydantic transforms input to floats, but the function fails unles integers are provided for NEW_SIZE
                width_param, height_param = round(width_param), round(height_param)

            resize_task = SpatialResizeTask(
                resize_type=self.config.resize.resize_type,
                height_param=height_param,
                width_param=width_param,
                resize_method=self.config.resize.method,
                resize_library=self.config.resize.library,
            )
            resize_node = EONode(resize_task, inputs=[import_node])

        save_task = SaveTask(
            self.storage.get_folder(self.config.output_folder_key),
            filesystem=self.storage.filesystem,
            compress_level=1,
            overwrite_permission=OverwritePermission.OVERWRITE_FEATURES,
            config=self.sh_config,
            features=[self.config.output_feature, FeatureType.BBOX],
        )
        save_node = EONode(save_task, inputs=[resize_node or import_node])

        return EOWorkflow.from_endnodes(save_node)



"""Implements a base training pipeline and a LGBM specialized classification and regression model training pipelines."""

import abc
import logging
from typing import Any, Dict, List, Literal, Optional, Tuple

import fs
import joblib
import numpy as np
from lightgbm import LGBMClassifier, LGBMRegressor
from pydantic import Field
from sklearn.metrics import (
    accuracy_score,
    f1_score,
    mean_absolute_error,
    mean_squared_error,
    precision_score,
    r2_score,
    recall_score,
)
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import LabelEncoder

from ..core.pipeline import Pipeline
from ..core.schemas import BaseSchema
from ..utils.validators import ensure_storage_key_presence

LOGGER = logging.getLogger(__name__)


class RandomTrainTestSplitSchema(BaseSchema):
    random_state: int = Field(
        42,
        description="Seed used in data splitter (either for scikit.learn.train_test_split or for scikit.utils.shuffle.",
    )
    # random training/testing split parameters
    train_size: float = Field(ge=0, le=1, description="Training size value (0.8 = 80/20 split for training/testing).")


class BaseTrainingPipeline(Pipeline, metaclass=abc.ABCMeta):
    """A base pipeline for training an ML model

    This class has a few abstract methods which have to be implemented. But in general all public methods are designed
    in a way that you can override them in a child class
    """

    class Schema(Pipeline.Schema):
        input_folder_key: str = Field(description="The storage manager key pointing to the model training data.")
        _ensure_input_folder_key = ensure_storage_key_presence("input_folder_key")
        model_folder_key: str = Field(
            description="The storage manager key pointing to the folder where the model will be saved."
        )
        _ensure_model_folder_key = ensure_storage_key_presence("model_folder_key")

        train_features: List[str] = Field(
            description="A list of feature filenames to join into training features in the given order."
        )

        train_reference: str = Field(description="Name of file where the reference data is stored.")

        train_test_split: RandomTrainTestSplitSchema

        model_parameters: Dict[str, Any] = Field(
            default_factory=dict, description="Parameters to be provided to the model"
        )
        model_filename: str
        patch_list: None = None
        input_patch_file: None = None
        skip_existing: Literal[False] = False

    config: Schema

    def run_procedure(self) -> Tuple[List[str], List[str]]:
        """The main pipeline procedure

        1. Prepares data. Output serves as input to both the training method and scoring method, so separation of
           training and testing data should be done within the object.
        2. Train model
        3. Save model
        4. Evaluate model
        """
        LOGGER.info("Preparing data")
        prepared_data = self.prepare_data()

        LOGGER.info("Training. This could take a while.")
        model = self.train_model(prepared_data)

        LOGGER.info("Saving.")
        self.save_model(model)

        LOGGER.info("Scoring results.")
        self.score_results(prepared_data, model)

        return [], []

    def prepare_data(self) -> dict:
        """Loads and preprocesses data."""
        features = self._collect_features()
        reference = self._collect_reference()

        features, reference = self.preprocess_data(features, reference)

        data_split = self.train_test_split(features, reference)
        prepared_data = dict(zip(["features_train", "features_test", "reference_train", "reference_test"], data_split))

        return prepared_data

    def _collect_features(self) -> np.ndarray:
        """Prepares features"""
        features = []
        LOGGER.info("Reading input features")

        for feature_name in self.config.train_features:
            array = self._read_array(feature_name)
            array = self._reshape_array(array)
            features.append(array)

        return np.concatenate(features, axis=1)

    def _collect_reference(self) -> np.ndarray:
        """Prepares reference"""
        LOGGER.info("Reading input reference")
        return self._read_array(self.config.train_reference)

    def preprocess_data(self, features: np.ndarray, reference: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """Preforms filtering and other preprocessing before splitting data."""
        return features, reference.ravel()

    def train_test_split(
        self, features: np.ndarray, reference: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Computes a random train-test split

        Order is train-features test-features train-reference test-reference.
        """
        config = self.config.train_test_split
        LOGGER.info("Making a random train-test split, using %s of data for training.", config.train_size)
        return train_test_split(features, reference, train_size=config.train_size, random_state=config.random_state)

    def _read_array(self, filename: str) -> np.ndarray:
        """Read numpy array from the training data folder given a filename."""
        folder = self.storage.get_folder(self.config.input_folder_key)
        if not filename.endswith(".npy"):
            filename = f"{filename}.npy"
        path = fs.path.combine(folder, filename)

        with self.storage.filesystem.openbin(path, "r") as file_handle:
            return np.load(file_handle)

    @staticmethod
    def _reshape_array(array: np.ndarray) -> np.ndarray:
        """Reshape numpy array into 2D representation suitable for fitting a model."""
        shape = array.shape
        return array.reshape((shape[0], np.prod(shape[1:])))

    @abc.abstractmethod
    def train_model(self, prepared_data: dict) -> object:
        """Trains the model on the data."""

    def save_model(self, model: object) -> None:
        """Saves the resulting model."""
        self._dump_object(self.config.model_filename, model)

    def _dump_object(self, filename: str, object_instance: object) -> None:
        """Dumps an object instance into models folder"""
        LOGGER.info("Saving to %s", filename)
        folder = self.storage.get_folder(self.config.model_folder_key)
        path = fs.path.combine(folder, filename)

        with self.storage.filesystem.openbin(path, "w") as file_handle:
            joblib.dump(object_instance, file_handle)

    @abc.abstractmethod
    def score_results(self, prepared_data: dict, model: Any) -> None:
        """Scores the resulting model and reports the metrics into the log files."""

    def predict(self, model: Any, features: np.ndarray) -> np.ndarray:  # pylint: disable=no-self-use
        """Evaluates model on features. Should be overridden for models with a different interface."""
        return model.predict(features)


class ClassificationPreprocessSchema(BaseSchema):
    label_encoder_filename: Optional[str] = Field(
        description="If specified uses a label encoder and saves it under specified name."
    )

    filter_classes: List[int] = Field(
        default_factory=list,
        description=(
            "Specify IDs of classes that are going to be used for training. If empty, all the classes will be used."
        ),
    )


class ClassificationTrainingPipeline(BaseTrainingPipeline):
    """A base pipeline for training an ML classifier. Uses LGBMClassifier by default."""

    class Schema(BaseTrainingPipeline.Schema):
        preprocessing: Optional[ClassificationPreprocessSchema]

    config: Schema

    def preprocess_data(self, features: np.ndarray, reference: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """Preforms filtering and other preprocessing before splitting data."""
        config = self.config.preprocessing
        reference = reference.ravel()
        if config is None:
            return features, reference

        if config.filter_classes:
            filtration = np.isin(reference, config.filter_classes).squeeze()
            features, reference = features[filtration, ...], reference[filtration]

        if config.label_encoder_filename:
            LOGGER.info("Applying label encoder to labels")
            label_encoder = LabelEncoder()
            label_encoder.fit(np.unique(reference))
            reference = label_encoder.transform(reference)

            self._dump_object(config.label_encoder_filename, label_encoder)

        return features, reference

    def train_model(self, prepared_data: dict) -> object:
        train_features = prepared_data["features_train"]
        train_reference = prepared_data["reference_train"]
        lgbm_classifier = LGBMClassifier(**self.config.model_parameters)
        lgbm_classifier.fit(train_features, train_reference)

        return lgbm_classifier

    def score_results(self, prepared_data: dict, model: Any) -> None:
        test_features = prepared_data["features_test"]
        test_reference = prepared_data["reference_test"]
        predictions = self.predict(model, test_features)

        accuracy = accuracy_score(test_reference, predictions)
        f1_value = f1_score(test_reference, predictions, average="weighted")
        recall = recall_score(test_reference, predictions, average="weighted")
        precision = precision_score(test_reference, predictions, average="weighted")

        LOGGER.info("Accuracy:\t %.4f", accuracy)
        LOGGER.info("F1 score:\t %.4f", f1_value)
        LOGGER.info("Recall:\t %.4f", recall)
        LOGGER.info("Precision:\t %.4f", precision)


class RegressionTrainingPipeline(BaseTrainingPipeline):
    """A base pipeline for training an ML regressor. Uses LGBMRegressor by default."""

    def train_model(self, prepared_data: dict) -> object:
        train_features = prepared_data["features_train"]
        train_reference = prepared_data["reference_train"]
        lgbm_classifier = LGBMRegressor(**self.config.model_parameters)
        lgbm_classifier.fit(train_features, train_reference)

        return lgbm_classifier

    def score_results(self, prepared_data: dict, model: Any) -> None:
        test_features = prepared_data["features_test"]
        test_reference = prepared_data["reference_test"]
        predictions = self.predict(model, test_features)

        r2_value = r2_score(test_reference, predictions)
        mse = mean_squared_error(test_reference, predictions)
        rmse = mean_squared_error(test_reference, predictions, squared=False)
        mae = mean_absolute_error(test_reference, predictions)

        LOGGER.info("Mean Square Error:\t %.4f", mse)
        LOGGER.info("Root Mean Square Error:\t %.4f", rmse)
        LOGGER.info("Mean Average Error:\t %.4f", mae)
        LOGGER.info("R2 Score:\t %.4f", r2_value)



"""Area manager implementation for automated UTM CRS grids."""

import logging
from collections import defaultdict
from typing import Dict, Optional

import fs
import geopandas as gpd
from geopandas import GeoDataFrame
from pydantic import Field

from sentinelhub import CRS, Geometry, UtmZoneSplitter

from ...utils.validators import field_validator
from ..schemas import BaseSchema
from .base import AreaSchema, BaseAreaManager, area_schema_deprecation, get_geometry_from_file

LOGGER = logging.getLogger(__name__)


class PatchSchema(BaseSchema):
    size_x: int = Field(description="A width of each EOPatch in meters")
    size_y: int = Field(description="A height of each EOPatch in meters")
    buffer_x: float = Field(0, description="Number of meters by which to increase the tile size to left and right.")
    buffer_y: float = Field(0, description="Number of meters by which to increase the tile size to up and down.")


class UtmZoneAreaManager(BaseAreaManager):
    """Area manager that splits the area per UTM zone"""

    class Schema(BaseAreaManager.Schema):
        area: Optional[AreaSchema] = Field(description="DEPRECATED, use `geometry_filename` instead.")
        geometry_filename: str = Field(  # type:ignore[assignment]
            None, description="Name of the file that defines the AoI geometry, located in the input data folder."
        )
        patch: PatchSchema

        offset_x: float = Field(0, description="An offset of tiling grid in horizontal dimension")
        offset_y: float = Field(0, description="An offset of tiling grid in vertical dimension")

        _warn_and_adapt_old_config = field_validator("geometry_filename", area_schema_deprecation, pre=True)

    config: Schema

    def get_area_geometry(self, *, crs: CRS = CRS.WGS84) -> Geometry:
        file_path = fs.path.join(self.storage.get_input_data_folder(), self.config.geometry_filename)
        return get_geometry_from_file(
            filesystem=self.storage.filesystem,
            file_path=file_path,
            geopandas_engine=self.storage.config.geopandas_backend,
        ).transform(crs)

    def _create_grid(self) -> Dict[CRS, GeoDataFrame]:
        """Uses UtmZoneSplitter to create a grid"""
        area_geometry = self.get_area_geometry()
        LOGGER.info("Splitting area geometry into UTM zone grid")
        splitter = UtmZoneSplitter(
            [area_geometry.geometry],
            crs=area_geometry.crs,
            bbox_size=(self.config.patch.size_x, self.config.patch.size_y),
            offset=(self.config.offset_x, self.config.offset_y),
        )

        bbox_list, info_list = splitter.get_bbox_list(), splitter.get_info_list()

        absolute_buffer = self.config.patch.buffer_x, self.config.patch.buffer_y
        if absolute_buffer != (0, 0):
            bbox_list = [bbox.buffer(absolute_buffer, relative=False) for bbox in bbox_list]

        crs_to_patches = defaultdict(list)
        zfill_length = len(str(len(bbox_list) - 1))
        for i, (bbox, info) in enumerate(zip(bbox_list, info_list)):
            i_x, i_y = info["index_x"], info["index_y"]
            name = f"eopatch-id-{i:0{zfill_length}}-col-{i_x}-row-{i_y}"
            crs_to_patches[bbox.crs].append((name, bbox.geometry))

        grid = {}
        for crs, named_bbox_geoms in crs_to_patches.items():
            names, geoms = zip(*named_bbox_geoms)
            grid[crs] = gpd.GeoDataFrame({self.NAME_COLUMN: names}, geometry=list(geoms), crs=crs.pyproj_crs())

        return grid

    def get_grid_cache_filename(self) -> str:
        input_filename = fs.path.basename(self.config.geometry_filename)
        input_filename = input_filename.rsplit(".", 1)[0]

        raw_params = [
            input_filename,
            self.config.patch.size_x,
            self.config.patch.size_y,
            self.config.patch.buffer_x,
            self.config.patch.buffer_y,
            self.config.offset_x,
            self.config.offset_y,
        ]
        params = [str(param) for param in raw_params]

        return f"{self.__class__.__name__}_{'_'.join(params)}.gpkg"



"""
Module containing utilities for working with vector data
"""
from typing import List, Optional, Union

import geopandas as gpd
import pandas as pd
from geopandas import GeoDataFrame
from shapely.geometry import MultiPolygon, Polygon

from sentinelhub import CRS


def concat_gdf(dataframe_list: List[GeoDataFrame], reproject_crs: Optional[CRS] = None) -> GeoDataFrame:
    """Concatenates together multiple GeoDataFrames, all in the same CRS

    There exists pandas.concat but no geopandas.concat. Therefore, this function implements it.

    :param dataframe_list: A list of GeoDataFrames to be concatenated together
    :param reproject_crs: A CRS in which dataframes should be reprojected before being joined
    :return: A joined GeoDataFrame
    """
    if not (dataframe_list and all(isinstance(gdf, GeoDataFrame) for gdf in dataframe_list)):
        raise ValueError(f"Expected a list of GeoDataFrames with at least 1 element, got {dataframe_list}")

    if reproject_crs is not None:
        crs = reproject_crs.pyproj_crs()
        dataframe_list = [gdf.to_crs(crs) for gdf in dataframe_list]
    else:
        unique_crs = {vector_gdf.crs for vector_gdf in dataframe_list}
        if len(unique_crs) > 1:
            raise ValueError("GeoDataFrames are in different CRS, therefore `reproject_crs` parameter should be given")

    return gpd.GeoDataFrame(pd.concat(dataframe_list, ignore_index=True), crs=dataframe_list[0].crs)


def count_points(geometry: Union[Polygon, MultiPolygon]) -> int:
    """Counts a number of points for a given geometry, both from exterior and interiors"""
    if isinstance(geometry, MultiPolygon):
        return sum(count_points(subgeometry) for subgeometry in geometry.geoms)
    exterior_count = len(geometry.exterior.coords)
    interior_count = sum(len(interior.coords) for interior in geometry.interiors)
    return exterior_count + interior_count



"""Utilities for working with lists of EOPatch names."""

import json
from collections import defaultdict
from typing import DefaultDict, Dict, List

from fs.base import FS

from sentinelhub import CRS

from ..types import PatchList


def save_names(filesystem: FS, file_path: str, names: List[str]) -> None:
    """Saves a list of names (EOPatch, execution, etc.) to a file

    :param filesystem: Filesystem used to save the file.
    :param filename: Path of a JSON file where names will be saved.
    :param names: A list of names.
    """

    with filesystem.open(file_path, "w") as file:
        json.dump(names, file, indent=2)


def load_names(filesystem: FS, file_path: str) -> List[str]:
    """Loads a list of names (EOPatch, execution, etc.) from a file

    :param filesystem: Filesystem used to load the file.
    :param filename: Path of a JSON file where names are saved.
    :return: A list of names loaded from file.
    """
    with filesystem.open(file_path, "r") as file:
        return json.load(file)


def group_by_crs(patch_list: PatchList) -> Dict[CRS, List[str]]:
    patches_by_crs: DefaultDict[CRS, List[str]] = defaultdict(list)
    for name, bbox in patch_list:
        patches_by_crs[bbox.crs].append(name)
    return patches_by_crs



"""
Modules with Ray-related utilities
"""
import logging
import os
import subprocess

import ray

from ..types import BoolOrAuto
from .general import current_timestamp

LOGGER = logging.getLogger(__name__)
CLUSTER_CONFIG_DIR = "~/.synced_configs"


def handle_ray_connection(use_ray: BoolOrAuto = "auto") -> bool:
    """According to the given parameter it will try to connect to an existing Ray cluster.

    :param use_ray: Either a boolean flag or `"auto"` to define if the connection should be established or not.
    :return: `True` if connection is established and `False` otherwise.
    """
    if use_ray == "auto":
        try:
            _try_connect_to_ray()
            return True
        except ConnectionError:
            LOGGER.info("No Ray cluster found, will not use Ray.")
            return False

    if use_ray:
        _try_connect_to_ray()
        return True
    return False


def _try_connect_to_ray() -> None:
    """Try connecting and log if successful."""
    ray.init(address="auto", ignore_reinit_error=True)
    LOGGER.info("Connected to an existing Ray cluster.")


def is_cluster_running(cluster_yaml: str) -> bool:
    """Checks if cluster is running or not."""
    try:
        subprocess.check_output(f"ray get_head_ip {cluster_yaml}", shell=True, stderr=subprocess.DEVNULL)
        return True
    except subprocess.CalledProcessError:
        return False


def start_cluster_if_needed(cluster_yaml: str) -> None:
    """Starts the cluster if it isn't already running."""
    if not is_cluster_running(cluster_yaml):
        subprocess.run(f"ray up {cluster_yaml} -y", shell=True)


def generate_cluster_config_path(config_filename: str) -> str:
    """Generate the path to the rsynced config on the cluster"""
    return f"{CLUSTER_CONFIG_DIR}/{current_timestamp()}{os.path.basename(config_filename)}"



"""
Module with utilities for creating maps
"""
import logging
import os
import shutil
import subprocess
import tempfile
import warnings
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Iterable, Literal, Optional

SH_COMMAND_LIMIT = 130000
OPEN_FILES_LIMIT = 1000


LOGGER = logging.getLogger(__name__)

GDAL_DTYPE_SETTINGS = {
    "uint8": "Byte",
    "uint16": "UInt16",
    "int16": "Int16",
    "float32": "Float32",
}
CogifyResamplingOptions = Literal[None, "NEAREST", "MODE", "AVERAGE", "BILINEAR", "CUBIC", "CUBICSPLINE", "LANCZOS"]
WarpResamplingOptions = Literal[
    None,
    "near",
    "bilinear",
    "cubic",
    "cubicspline",
    "lanczos",
    "average",
    "rms",
    "mode",
    "max",
    "min",
    "med",
    "q1",
    "q3",
    "sum",
]


def cogify_inplace(
    tiff_file: str,
    blocksize: int = 2048,
    nodata: Optional[float] = None,
    dtype: Literal[None, "int8", "int16", "uint8", "uint16", "float32"] = None,
    resampling: CogifyResamplingOptions = None,
    quiet: bool = True,
) -> None:
    """Make the (geotiff) file a cog

    :param tiff_file: .tiff file to cogify
    :param blocksize: block size of tiled COG
    :param nodata: value to be treated as nodata, default value is None
    :param dtype: output type of the in the resulting tiff, default is None
    :param resampling: The resampling method used to produce overviews. The defaults (when using None) are CUBIC for
        floats and NEAREST for integers.
    :param quiet: The process does not produce logs.
    """
    temp_file = NamedTemporaryFile()
    temp_file.close()

    cogify(
        tiff_file,
        temp_file.name,
        blocksize,
        nodata=nodata,
        dtype=dtype,
        overwrite=True,
        resampling=resampling,
        quiet=quiet,
    )
    # Note: by moving the file we also remove the one at temp_file.name
    shutil.move(temp_file.name, tiff_file)


def cogify(
    input_file: str,
    output_file: str,
    blocksize: int = 1024,
    nodata: Optional[float] = None,
    dtype: Literal[None, "int8", "int16", "uint8", "uint16", "float32"] = None,
    overwrite: bool = True,
    resampling: CogifyResamplingOptions = None,
    quiet: bool = True,
) -> None:
    """Create a cloud optimized version of input file

    :param input_file: File to cogify
    :param output_file: Resulting cog file
    :param blocksize: block size of tiled COG
    :param nodata: value to be treated as nodata, default value is None
    :param dtype: output type of the in the resulting tiff, default is None
    :param overwrite: If True overwrite the output file if it exists.
    :param resampling: The resampling method used to produce overviews. The defaults (when using None) are CUBIC for
        floats and NEAREST for integers.
    :param quiet: The process does not produce logs.
    """
    if input_file == output_file:
        raise OSError("Input file is the same as output file.")

    if os.path.exists(output_file):
        if overwrite:
            os.remove(output_file)
        else:
            raise OSError(f"{output_file} exists!")

    version = subprocess.check_output(("gdalinfo", "--version"), text=True).split(",")[0].replace("GDAL ", "")
    if version < "3.1.0":
        raise RuntimeError(
            f"The cogification process is configured for GDAL 3.1.0 and higher, but version {version} was detected.",
            RuntimeWarning,
        )

    gdaltranslate_options = (
        f"-of COG -co COMPRESS=DEFLATE -co BLOCKSIZE={blocksize} -co OVERVIEWS=IGNORE_EXISTING -co PREDICTOR=YES"
    )

    if resampling:
        gdaltranslate_options += f" -co RESAMPLING={resampling}"

    if quiet:
        gdaltranslate_options += " -q"

    if nodata is not None:
        gdaltranslate_options += f" -a_nodata {nodata}"

    if dtype is not None:
        gdaltranslate_options += f" -ot {GDAL_DTYPE_SETTINGS[dtype]}"

    if version < "3.6.0" and resampling == "MODE":
        warnings.warn(
            (
                "GDAL versions below 3.6.0 have issues with `MODE` overview resampling. Trying to fix issue by setting"
                " GDAL_OVR_CHUNK_MAX_SIZE to a large integer (2100000000)."
            ),
            category=RuntimeWarning,
            stacklevel=2,
        )
        gdaltranslate_options += " --config GDAL_OVR_CHUNK_MAX_SIZE 2100000000"

    subprocess.check_call(f"gdal_translate {gdaltranslate_options} {input_file} {output_file}", shell=True)


def merge_tiffs(
    input_filenames: Iterable[str],
    merged_filename: str,
    *,
    overwrite: bool = True,
    nodata: Optional[float] = None,
    dtype: Literal[None, "int8", "int16", "uint8", "uint16", "float32"] = None,
    warp_resampling: WarpResamplingOptions = None,
    quiet: bool = True,
) -> None:
    """Performs gdal_merge on a set of given geotiff images

    :param input_filenames: A sequence of input tiff image filenames
    :param merged_filename: Filename of merged tiff image
    :param overwrite: If True overwrite the output (merged) file if it exists
    :param delete_input: If True input images will be deleted at the end
    :param warp_resampling: The resampling method used when warping, useful for pixel misalignment. Defaults to NEAREST.
    :param quiet: The process does not produce logs.
    """
    gdalwarp_options = "-co BIGTIFF=YES -co compress=LZW -co TILED=YES"

    if overwrite:
        gdalwarp_options += " -overwrite"

    if quiet:
        gdalwarp_options += " -q"

    if warp_resampling:
        gdalwarp_options += f" -r {warp_resampling}"

    if nodata is not None:
        gdalwarp_options += f' -dstnodata "{nodata}"'  # noqa B028

    if dtype is not None:
        gdalwarp_options += f" -ot {GDAL_DTYPE_SETTINGS[dtype]}"

    input_filelist = list(input_filenames)
    command = f"gdalwarp {gdalwarp_options} {' '.join(input_filelist)} {merged_filename}"

    if len(command) > SH_COMMAND_LIMIT or len(input_filelist) > OPEN_FILES_LIMIT:
        merged_path = Path(merged_filename)
        vrt_file_path = merged_path.with_name(f"{merged_path.stem}_temp.vrt")
        LOGGER.info(f"Command too big or too many files to process. Creating an intermediary vrt: {vrt_file_path}")
        # generate text file with tile names & generate vrt
        with tempfile.NamedTemporaryFile(mode="w", delete=True) as file_list:
            file_list.writelines([f"{tname}\n" for tname in input_filelist])
            generate_vrt_command = f"gdalbuildvrt {vrt_file_path} -input_file_list {file_list.name}"
            subprocess.check_call(generate_vrt_command, shell=True)

        # create merged file
        command = f"gdalwarp {gdalwarp_options} {vrt_file_path} {merged_filename}"
        subprocess.check_call(command, shell=True)

        # cleanup the vrt after the process
        os.remove(vrt_file_path)
    else:
        subprocess.check_call(command, shell=True)


def extract_bands(
    input_file: str,
    output_file: str,
    bands: Iterable[int],
    overwrite: bool = True,
    compress: bool = False,
    quiet: bool = True,
) -> None:
    """Extract bands from given input file

    :param input_file: File containing all bands
    :param output_file: Resulting file with extracted bands
    :param bands: Sequence of bands to extract. Indexation starts at 0.
    :param overwrite: If True overwrite the output file if it exists.
    :param quiet: The process does not produce logs.
    """
    if not bands:
        raise ValueError("No bands were specified for extraction, undefined behaviour.")

    if input_file == output_file:
        raise OSError("Input file is the same as output file.")

    if os.path.exists(output_file):
        if overwrite:
            os.remove(output_file)
        else:
            raise OSError(f"{output_file} already exists. Set `overwrite` to true if it should be overwritten.")

    # gdal_translate starts indexation at 1
    translate_opts = " ".join(f" -b {band + 1}" for band in bands)
    if quiet:
        translate_opts += " -q"
    if compress:
        translate_opts += " -co compress=LZW"

    command = f"gdal_translate {translate_opts} {input_file} {output_file}"
    subprocess.check_call(command, shell=True)



"""Implements different pipelines for sampling from data."""
import abc
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
from pydantic import Field

from eolearn.core import EONode, EOWorkflow, FeatureType, LoadTask, MergeEOPatchesTask, OverwritePermission, SaveTask
from eolearn.geometry import MorphologicalOperations, MorphologicalStructFactory
from eolearn.ml_tools import BlockSamplingTask, FractionSamplingTask, GridSamplingTask

from eogrow.utils.validators import ensure_exactly_one_defined, ensure_storage_key_presence

from ..core.pipeline import Pipeline
from ..tasks.common import ClassFilterTask
from ..types import ExecKwargs, Feature, FeatureSpec, PatchList
from ..utils.filter import get_patches_with_missing_features


class BaseSamplingPipeline(Pipeline, metaclass=abc.ABCMeta):
    """Pipeline to run sampling on EOPatches"""

    class Schema(Pipeline.Schema):
        output_folder_key: str = Field(description="The storage manager key pointing to the pipeline output folder.")
        _ensure_output_folder_key = ensure_storage_key_presence("output_folder_key")

        apply_to: Dict[str, Dict[FeatureType, List[str]]] = Field(
            description=(
                "A dictionary defining which features to sample, its structure is "
                "{folder_key: {feature_type: [feature_name]}}"
            ),
        )
        mask_of_samples_name: Optional[str] = Field(
            description=(
                "A name of a mask timeless output feature with information which pixels were sampled and how many times"
            )
        )
        sampled_suffix: Optional[str] = Field(
            description=(
                "If provided features are saved with a suffix, e.g. for suffix SAMPLED the sampled FEATURES are "
                "saved as FEATURES_SAMPLED."
            )
        )
        compress_level: int = Field(1, description="Level of compression used in saving eopatches")

    config: Schema

    def filter_patch_list(self, patch_list: PatchList) -> PatchList:
        """Filter output EOPatches that have already been processed"""
        filtered_patch_list = get_patches_with_missing_features(
            self.storage.filesystem,
            self.storage.get_folder(self.config.output_folder_key),
            patch_list,
            self._get_output_features(),
        )

        return filtered_patch_list

    def build_workflow(self) -> EOWorkflow:
        """Creates workflow that is divided into the following sub-parts:

        1. loading data,
        2. preprocessing steps,
        3. sampling features
        4. saving results
        """
        loading_node = self._get_loading_node()
        preprocessing_node = self._get_preprocessing_node(loading_node)
        sampling_node = self._get_sampling_node(preprocessing_node)

        save_task = SaveTask(
            self.storage.get_folder(self.config.output_folder_key),
            filesystem=self.storage.filesystem,
            features=self._get_output_features(),
            compress_level=self.config.compress_level,
            overwrite_permission=OverwritePermission.OVERWRITE_FEATURES,
        )

        return EOWorkflow.from_endnodes(EONode(save_task, inputs=[sampling_node]))

    def _get_loading_node(self) -> EONode:
        """Prepares nodes for loading and joining EOPatches."""
        load_nodes = []

        for folder_name, features in self.config.apply_to.items():
            load_features: List[FeatureSpec] = []

            for feature_type_str, feature_names in features.items():
                feature_type = FeatureType(feature_type_str)

                if not feature_type.is_spatial():
                    raise ValueError(f"Only spatial features can be sampled, but found {feature_type}: {feature_names}")

                for feature_name in feature_names:
                    load_features.append((feature_type, feature_name))

            load_features.append(FeatureType.BBOX)
            if any(FeatureType(feature_type).is_temporal() for feature_type in features):
                load_features.append(FeatureType.TIMESTAMPS)

            load_task = LoadTask(
                self.storage.get_folder(folder_name),
                filesystem=self.storage.filesystem,
                lazy_loading=True,
                features=load_features,
            )
            load_nodes.append(EONode(load_task, name=f"Load from {folder_name}"))

        return EONode(MergeEOPatchesTask(), inputs=load_nodes)

    def _get_preprocessing_node(self, previous_node: EONode) -> EONode:  # pylint: disable=no-self-use
        """The default implementation doesn't add any preprocessing steps"""
        return previous_node

    @abc.abstractmethod
    def _get_sampling_node(self, previous_node: EONode) -> EONode:
        """Method to prepare sampling nodes"""

    def _get_features_to_sample(self) -> List[Tuple[FeatureType, str, str]]:
        """Get a list of features that will be sampled, together with their new names"""
        features_to_sample = []
        for _, features in self.config.apply_to.items():
            for feature_type, feature_names in features.items():
                for feature_name in feature_names:
                    if self.config.sampled_suffix is None:
                        features_to_sample.append((feature_type, feature_name, feature_name))
                    else:
                        features_to_sample.append(
                            (feature_type, feature_name, f"{feature_name}_{self.config.sampled_suffix}")
                        )

        return features_to_sample

    def _get_mask_of_samples_feature(self) -> Optional[Feature]:
        """Provide a mask of samples feature"""
        if self.config.mask_of_samples_name:
            return FeatureType.MASK_TIMELESS, self.config.mask_of_samples_name
        return None

    def _get_output_features(self) -> List[FeatureSpec]:
        """Get a list of features that will be saved as an output of the pipeline"""
        output_features: List[FeatureSpec] = [FeatureType.BBOX]
        features_to_sample = self._get_features_to_sample()

        for feature_type, _, sampled_feature_name in features_to_sample:
            output_features.append((feature_type, sampled_feature_name))

        mask_of_samples_feature = self._get_mask_of_samples_feature()
        if mask_of_samples_feature:
            output_features.append(mask_of_samples_feature)

        if any(feature_type.is_temporal() for feature_type, _, _ in features_to_sample):
            output_features.append(FeatureType.TIMESTAMPS)
        return output_features


class BaseRandomSamplingPipeline(BaseSamplingPipeline, metaclass=abc.ABCMeta):  # noqa B024
    """A base class for all sampling pipeline that work on random selection of samples"""

    class Schema(BaseSamplingPipeline.Schema):
        seed: Optional[int] = Field(
            42, description="A random generator seed to be used in order to obtain the same results every pipeline run."
        )

    config: Schema

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self._sampling_node_uid: Optional[str] = None

    def get_execution_arguments(self, workflow: EOWorkflow, patch_list: PatchList) -> ExecKwargs:
        """Extends the basic method for adding execution arguments by adding seed arguments a sampling task"""
        exec_args = super().get_execution_arguments(workflow, patch_list)

        sampling_node = workflow.get_node_with_uid(self._sampling_node_uid)
        if sampling_node is None:
            return exec_args

        generator = np.random.default_rng(seed=self.config.seed)

        for _, patch_args in exec_args.items():
            patch_args[sampling_node] = dict(seed=generator.integers(low=0, high=2**32))

        return exec_args


class FractionSamplingPipeline(BaseRandomSamplingPipeline):
    """A pipeline to sample per-class with different distributions"""

    class Schema(BaseRandomSamplingPipeline.Schema):
        sampling_feature_name: str = Field(
            description="Name of MASK_TIMELESS feature to be used to create sample point"
        )
        erosion_dict: Optional[Dict[int, List[int]]] = Field(
            description="A dictionary specifying disc radius of erosion operation to be applied to a list of label IDs",
            example={2: [1, 3, 4], 1: [2]},
        )
        fraction_of_samples: Union[float, Dict[int, float]] = Field(
            description=(
                "A fraction or a dictionary of per-class fractions of valid pixels to sample from the the sampling "
                "feature."
            )
        )
        exclude_values: List[int] = Field(default_factory=list, description="Values to be excluded from sampling")

    config: Schema

    def _get_preprocessing_node(self, previous_node: EONode) -> EONode:
        """Preprocessing that applies erosion on sampling feature values"""
        if self.config.erosion_dict is None:
            return previous_node

        end_node = previous_node
        for radius, labels in self.config.erosion_dict.items():
            task = ClassFilterTask(
                (FeatureType.MASK_TIMELESS, self.config.sampling_feature_name),
                labels,
                MorphologicalOperations.EROSION,
                struct_elem=MorphologicalStructFactory.get_disk(radius),
            )
            end_node = EONode(task, inputs=[end_node])

        return end_node

    def _get_sampling_node(self, previous_node: EONode) -> EONode:
        """Prepare the sampling task"""
        task = FractionSamplingTask(
            features_to_sample=self._get_features_to_sample(),
            sampling_feature=(FeatureType.MASK_TIMELESS, self.config.sampling_feature_name),
            fraction=self.config.fraction_of_samples,
            exclude_values=self.config.exclude_values,
            mask_of_samples=self._get_mask_of_samples_feature(),
        )
        node = EONode(task, inputs=[previous_node])
        self._sampling_node_uid = node.uid

        return node


class BlockSamplingPipeline(BaseRandomSamplingPipeline):
    """A pipeline to randomly sample blocks"""

    class Schema(BaseRandomSamplingPipeline.Schema):
        sample_size: Tuple[int, int] = Field(description="A height and width of each block in pixels.")

        number_of_samples: Optional[int] = Field(
            description=(
                "A number of samples to be sampled. Exactly one of parameters fraction_of_samples and number_of_samples"
                " has to be given."
            )
        )
        fraction_of_samples: Optional[float] = Field(
            description=(
                "A percentage of samples to be sampled. Exactly one of parameters fraction_of_samples and "
                "number_of_samples has to be given."
            )
        )

        _check_fraction_number = ensure_exactly_one_defined("number_of_samples", "fraction_of_samples")

    config: Schema

    def _get_sampling_node(self, previous_node: EONode) -> EONode:
        """Prepare the sampling task"""
        amount: float = self.config.fraction_of_samples or self.config.number_of_samples  # type: ignore[assignment]
        task = BlockSamplingTask(
            features_to_sample=self._get_features_to_sample(),
            amount=amount,
            sample_size=self.config.sample_size,
            mask_of_samples=self._get_mask_of_samples_feature(),
        )

        node = EONode(task, inputs=[previous_node])
        self._sampling_node_uid = node.uid

        return node


class GridSamplingPipeline(BaseSamplingPipeline):
    """A pipeline to sample blocks in a regular grid"""

    class Schema(BaseSamplingPipeline.Schema):
        sample_size: Tuple[int, int] = Field(description="A height and width of each block in pixels.")
        stride: Tuple[int, int] = Field(
            description=(
                "A tuple describing a distance between upper left corners of two consecutive sampled blocks. "
                "The first number is the vertical distance and the second number the horizontal distance. If stride in "
                "smaller than sample_size in any dimensions then sampled blocks will overlap."
            )
        )

    config: Schema

    def _get_sampling_node(self, previous_node: EONode) -> EONode:
        """Prepare the sampling task"""
        task = GridSamplingTask(
            features_to_sample=self._get_features_to_sample(),
            sample_size=self.config.sample_size,
            stride=self.config.stride,
            mask_of_samples=self._get_mask_of_samples_feature(),
        )

        return EONode(task, inputs=[previous_node])



"""Implements functions that transform raw dictionaries/JSON files according to the config language of eo-grow."""
import copy
import os
import re
from functools import reduce
from typing import Any, Callable, Dict, List, NewType, Optional, Set, cast

import fs.path
import rapidjson

from eolearn.core.utils.fs import get_base_filesystem_and_path, get_full_path, join_path

from ..utils.meta import get_os_import_path

CrudeConfig = NewType("CrudeConfig", dict)
RawConfig = NewType("RawConfig", dict)


def collect_configs_from_path(path: str, used_config_paths: Optional[Set[str]] = None) -> List[CrudeConfig]:
    """Loads and builds a list of config dictionaries defined by the parameters stored in files

    This function performs the 1st stage of language interpretation as described in
    eo-grow/documentation/config-language.md`.

    :param path: A full path where a config file is located
    :param used_config_paths: A helper parameter to prevent cyclic config imports.
    :return: A list of stage 1 dictionaries from which to build config object.
    """
    filesystem, path = get_base_filesystem_and_path(path)
    with filesystem.open(path, "r") as file_handle:
        config = rapidjson.load(file_handle, parse_mode=rapidjson.PM_COMMENTS)

    used_config_paths = copy.copy(used_config_paths) or set()
    used_config_paths.add(get_full_path(filesystem, path))

    full_folder_path = get_full_path(filesystem, fs.path.dirname(path))
    config = _recursive_apply_to_strings(config, lambda value: _resolve_config_paths(value, full_folder_path))

    config = _recursive_config_build(config, used_config_paths)

    if isinstance(config, dict):
        config = [config]
    if isinstance(config, list):
        return config
    raise ValueError(f"When interpreting config from {path} a dictionary or list was expected, got {type(config)}.")


def _recursive_config_build(config: object, used_config_paths: Set[str]) -> object:
    """Recursively builds a configuration object by replacing dictionary items in form of

        `'**key': 'file path to another config'`

    with a content of a configuration file, referenced with the path. The items from both configuration objects are
    joined recursively.
    """
    if isinstance(config, dict):
        joint_config = {}
        imported_configs: List[CrudeConfig] = []

        for key, value in config.items():
            if not isinstance(key, str):
                raise ValueError(f"Dictionary keys should always be strings, but found: {key}")

            if key.startswith("**"):
                if value in used_config_paths:
                    raise ValueError("Detected a cyclic import of configs")

                imported_config_list = collect_configs_from_path(value, used_config_paths=used_config_paths)
                if len(imported_config_list) != 1:
                    raise ValueError(
                        "Config lists cannot be imported inside configs. Found a config list when resolving key"
                        f" {key} for path {value}"
                    )
                imported_configs.append(imported_config_list[0])
            else:
                joint_config[key] = _recursive_config_build(value, used_config_paths)

        return reduce(recursive_config_join, imported_configs, joint_config)

    if isinstance(config, list):
        return [_recursive_config_build(value, used_config_paths) for value in config]

    return config


def interpret_config_from_dict(config: CrudeConfig, external_variables: Optional[Dict[str, Any]] = None) -> RawConfig:
    """Applies config language rules to a loaded config

    This function performs the 2nd stage of language interpretation as described in
    `eo-grow/documentation/config-language.md`.
    """
    _recursive_check_config(config)
    config_with_imports = _recursive_apply_to_strings(config, _resolve_import_paths)
    if not isinstance(config_with_imports, dict):
        raise ValueError(f"Interpretation resulted in object of type {type(config)} but a dictionary was expected.")
    config = cast(CrudeConfig, config_with_imports)

    variable_mapping = config.pop("variables", {})
    if external_variables:
        variable_mapping.update(external_variables)

    for variable_name in variable_mapping:
        if not re.fullmatch(r"\w+", variable_name):
            raise ValueError(f"Variable name {variable_name} contains illegal characters")

    config_with_variables = _recursive_apply_to_strings(
        config, lambda config_str: _resolve_variables(config_str, variable_mapping)
    )
    if not isinstance(config_with_variables, dict):
        raise ValueError(
            f"Interpretation resulted in object of type {type(config_with_variables)} but a dictionary was expected."
        )

    return cast(RawConfig, config_with_variables)


def interpret_config_from_path(path: str) -> RawConfig:
    """Loads from path in applies both steps of the config language."""
    configs = collect_configs_from_path(path)
    if len(configs) != 1:
        raise ValueError(f"The JSON file {path} was expected to contain a single dictionary, got {len(configs)}")
    return interpret_config_from_dict(configs[0])


def _resolve_config_paths(config_str: str, config_path: str) -> str:
    """Replaces `${config_path}` with an actual path"""
    new_config_str = re.sub(r"\${config_path}", config_path, config_str)

    if new_config_str != config_str:
        return join_path(new_config_str)
    return new_config_str


def _resolve_import_paths(config_str: str) -> str:
    """Replaces `${import_path:package.module}` with an actual local path"""
    new_config_str = re.sub(r"\${import_path:([\w.]+)}", _sub_import_path, config_str)

    if new_config_str != config_str:
        return join_path(new_config_str)
    return new_config_str


def _sub_import_path(match: re.Match) -> str:
    """Substitutes a regex match with a local file path to a module or a package"""
    import_module = match.group(1)
    return os.path.dirname(get_os_import_path(import_module))


def _resolve_variables(config_str: str, variable_mapping: Dict[str, str]) -> str:
    """Replaces `${var:variable_name}` with a replacement value defined under variables"""
    return re.sub(r"\${var:(\w+)}", lambda match: _sub_variable(match, variable_mapping), config_str)


def _sub_variable(match: re.Match, variable_mapping: Dict[str, str]) -> str:
    """Substitutes a regex match with a new variable according to the mapping"""
    variable_name = match.group(1)
    if variable_name in variable_mapping:
        return str(variable_mapping[variable_name])
    raise ValueError(f"Variable name `{variable_name}` doesn't exist in a config dictionary of variables.")


def _recursive_apply_to_strings(config: object, function: Callable) -> object:
    """Recursively applies a function on all string values (and not keys) of a nested config object"""
    if isinstance(config, dict):
        return {key: _recursive_apply_to_strings(value, function) for key, value in config.items()}

    if isinstance(config, list):
        return [_recursive_apply_to_strings(value, function) for value in config]

    if isinstance(config, str):
        return function(config)
    return config


def _recursive_check_config(config: object) -> None:
    """Recursively checks if the config satisfies basic conditions for being JSON serializable.

    :raises: ValueError
    """
    if isinstance(config, dict):
        for key, value in config.items():
            if not isinstance(key, str):
                raise ValueError(f"Config keys should be strings but {key} found")
            _recursive_check_config(value)

    elif isinstance(config, list):
        for value in config:
            _recursive_check_config(value)


def recursive_config_join(config1: dict, config2: dict) -> dict:
    """Recursively join 2 config objects, where config1 values override config2 values"""
    for key, value in config2.items():
        if key not in config1:
            config1[key] = value
        elif isinstance(config1[key], dict) and isinstance(value, dict):
            config1[key] = recursive_config_join(config1[key], value)

    return config1



"""Base object from which all configurable eo-grow objects inherit."""
from typing import Any, Type, TypeVar

from pydantic import BaseModel

from .config import RawConfig, interpret_config_from_path

Self = TypeVar("Self", bound="EOGrowObject")


class EOGrowObject:
    """A base object in `eo-grow` framework"""

    class Schema(BaseModel):
        """A pydantic parsing/validation schema describing the shape of input parameters."""

        class Config:
            """Forbids unspecified fields and validates default values as well."""

            extra = "forbid"
            validate_all = True
            allow_mutation = False
            arbitrary_types_allowed = True

    config: Schema

    def __init__(self, config: Schema):
        self.config = config

    @classmethod
    def from_raw_config(cls: Type[Self], config: RawConfig, *args: Any, **kwargs: Any) -> Self:
        """Creates an object from a dictionary by constructing a validated config and use it to create the object."""
        validated_config = cls.Schema.parse_obj(config)
        return cls(validated_config, *args, **kwargs)

    @classmethod
    def from_path(cls: Type[Self], path: str, *args: Any, **kwargs: Any) -> Self:
        """Creates an object by loading and validating a config from a JSON file."""
        config = interpret_config_from_path(path)
        return cls.from_raw_config(config, *args, **kwargs)



"""Area manager implementation for Sentinel Hub batch grids."""
import logging
from collections import defaultdict
from typing import Dict, Optional

import fs
import geopandas as gpd
from geopandas import GeoDataFrame
from pydantic import Field

from sentinelhub import CRS, BatchRequest, BatchRequestStatus, BatchSplitter, Geometry, SentinelHubBatch

from ...utils.validators import field_validator
from ..storage import StorageManager
from .base import AreaSchema, BaseAreaManager, area_schema_deprecation, get_geometry_from_file

LOGGER = logging.getLogger(__name__)


class MissingBatchIdError(ValueError):
    """Exception that is triggered if ID of a Sentinel Hub batch job is missing."""


class BatchAreaManager(BaseAreaManager):
    """Area manager that splits the area according to the Sentinel Hub Batch tiling grids."""

    class Schema(BaseAreaManager.Schema):
        area: Optional[AreaSchema] = Field(description="DEPRECATED, use `geometry_filename` instead.")
        geometry_filename: str = Field(  # type:ignore[assignment]
            None, description="Name of the file that defines the AoI geometry, located in the input data folder."
        )
        tiling_grid_id: int = Field(
            description="An id of one of the tiling grids predefined at Sentinel Hub Batch service."
        )
        resolution: float = Field(
            description=(
                "Resolution of downloaded imagery in meters. Resolution options are predefined at Sentinel Hub Batch"
                " service for a chosen tiling_grid_id."
            )
        )
        tile_buffer_x: int = Field(0, description="Number of pixels for which to buffer each tile left and right.")
        tile_buffer_y: int = Field(0, description="Number of pixels for which to buffer each tile up and down.")
        batch_id: Optional[str] = Field(
            description=(
                "ID of the batch job that defines the AOI. Not required when using BatchDownloadPipeline,"
                " which generates a new batch job with the given AOI parameters."
            ),
        )

        _warn_and_adapt_old_config = field_validator("geometry_filename", area_schema_deprecation, pre=True)

    config: Schema

    def __init__(self, config: Schema, storage: StorageManager):
        super().__init__(config, storage)
        # We provide a way to inject a Batch ID after initialization if no ID was given in the config
        # This is meant to be used only in the BatchDownloadPipeline to force caching
        self._injected_batch_id: Optional[str] = None

    def get_area_geometry(self, *, crs: CRS = CRS.WGS84) -> Geometry:
        file_path = fs.path.join(self.storage.get_input_data_folder(), self.config.geometry_filename)
        return get_geometry_from_file(
            filesystem=self.storage.filesystem,
            file_path=file_path,
            geopandas_engine=self.storage.config.geopandas_backend,
        ).transform(crs)

    def _create_grid(self) -> Dict[CRS, GeoDataFrame]:
        """Uses BatchSplitter to create a grid for the selected batch job."""
        batch_id = self.config.batch_id or self._injected_batch_id

        if batch_id is None:
            raise MissingBatchIdError(
                "Trying to create a new batch grid but cannot collect tile geometries because 'batch_id' has not been "
                f"given. You can either provide it in the {self.__class__.__name__} schema or run a pipeline that "
                "creates a new batch request."
            )

        batch_client = SentinelHubBatch(config=self.storage.sh_config)
        batch_request = batch_client.get_request(batch_id)
        self._verify_batch_request(batch_request)

        splitter = BatchSplitter(batch_request=batch_request, config=self.storage.sh_config)
        bbox_list, info_list = splitter.get_bbox_list(), splitter.get_info_list()

        crs_to_patches = defaultdict(list)
        # they are returned in random order, so we sort them by name beforehand
        for bbox, info in sorted(zip(bbox_list, info_list), key=lambda x: x[1]["name"]):  # type: ignore
            crs_to_patches[bbox.crs].append((info["name"], bbox.geometry))

        grid = {}
        for crs, named_bbox_geoms in crs_to_patches.items():
            names, geoms = zip(*named_bbox_geoms)
            grid[crs] = gpd.GeoDataFrame({self.NAME_COLUMN: names}, geometry=list(geoms), crs=crs.pyproj_crs())

        return grid

    def _verify_batch_request(self, batch_request: BatchRequest) -> None:
        """Verifies that the given batch request has finished and that it contains the same tiling grid parameters as
        in the config.
        """
        batch_request.raise_for_status(
            status=(
                BatchRequestStatus.CREATED,  # tiles not available prior to analysis
                BatchRequestStatus.ANALYSING,  # tiles not available prior to analysis
                BatchRequestStatus.FAILED,
                BatchRequestStatus.CANCELED,
            )
        )

        expected_tiling_grid_params = {
            "id": self.config.tiling_grid_id,
            "bufferY": self.config.tile_buffer_y,
            "bufferX": self.config.tile_buffer_x,
            "resolution": self.config.resolution,
        }
        if batch_request.tiling_grid != expected_tiling_grid_params:
            raise ValueError(
                f"Tiling grid parameters in config are {expected_tiling_grid_params} but given batch "
                f"request has parameters {batch_request.tiling_grid}"
            )

    def get_grid_cache_filename(self) -> str:
        input_filename = fs.path.basename(self.config.geometry_filename)
        input_filename = input_filename.rsplit(".", 1)[0]

        raw_params = [
            input_filename,
            self.config.tiling_grid_id,
            self.config.resolution,
            self.config.tile_buffer_x,
            self.config.tile_buffer_y,
        ]
        params = [str(param) for param in raw_params]

        return f"{self.__class__.__name__}_{'_'.join(params)}.gpkg"



"""Implementation of the base AreaManager."""
import logging
import warnings
from abc import ABCMeta, abstractmethod
from typing import Dict, Literal, Optional

import fiona
import fs
import geopandas as gpd
import shapely.ops
from pydantic import Field

from eolearn.core.exceptions import EODeprecationWarning
from sentinelhub import CRS, BBox, Geometry

from ...types import PatchList, RawSchemaDict
from ...utils.eopatch_list import load_names
from ...utils.fs import LocalFile
from ..base import EOGrowObject
from ..schemas import BaseSchema, ManagerSchema
from ..storage import StorageManager

LOGGER = logging.getLogger(__name__)


class AreaSchema(BaseSchema):
    filename: str
    buffer: Optional[float] = Field(
        description="Buffer that will be applied to AOI geometry. Buffer has to be in the same units as AOI CRS.",
    )
    simplification_factor: Optional[float] = Field(
        description="Tolerance factor (in CRS units) for simplifying the buffered area geometry before splitting it.",
    )


def area_schema_deprecation(cls: type, value: Optional[str], values: RawSchemaDict) -> str:
    """Warns and reconfigures when `area` is used instead of `geometry_filename`."""
    if values.get("area") is not None:
        warnings.warn(
            (
                "Use `geometry_filename` to provide the file (e.g., geojson, gpkg) with AoI. The `area` parameter is"
                " deprecated, and extra parameters like `buffer` and `simplification_factor` are no longer available;"
                " users should prepare the AoI geometry by themselves."
            ),
            EODeprecationWarning,
            stacklevel=2,
        )
        return values["area"].filename
    assert value is not None, "Specify the `geometry_filename` parameter."
    return value


class PatchListSchema(BaseSchema):
    input_folder_key: str = Field(description="The storage manager key pointing to the folder containing the file.")
    filename: str = Field(description="A JSON file containing a list of EOPatch names.", regex=r"^.+\.(json|JSON)$")


class BaseAreaManager(EOGrowObject, metaclass=ABCMeta):
    """A manager for the AOI and how it is split into EOPatches"""

    NAME_COLUMN = "eopatch_name"

    class Schema(ManagerSchema):
        patch_names: Optional[PatchListSchema] = Field(
            description="Names of EOPatches to keep when filtering in the `get_grid` method."
        )

    config: Schema

    def __init__(self, config: Schema, storage: StorageManager):
        """
        :param config: The configuration schema
        :param storage: An instance of StorageManager class
        """
        super().__init__(config)

        self.storage = storage

    @abstractmethod
    def get_area_geometry(self, *, crs: CRS = CRS.WGS84) -> Geometry:
        """Provides a dissolved geometry object of the entire AOI"""

    def get_grid(self, filtered: bool = True) -> Dict[CRS, gpd.GeoDataFrame]:
        """Provides a grid of bounding boxes which divide the AOI. Uses caching to avoid recalculations.

        The grid is split into different CRS zones. The `bounds` properties of the geometries are taken as BBox
        definitions. EOPatch names are stored in a column with identifier `self.NAME_COLUMN`.

        :return: A dictionary of GeoDataFrames that defines how the area is split into EOPatches.
        """
        grid_path = fs.path.combine(self.storage.get_cache_folder(), self.get_grid_cache_filename())

        if self.storage.filesystem.exists(grid_path):
            grid = self._load_grid(grid_path)
        else:
            grid = self._create_grid()
            self._save_grid(grid, grid_path)

        if filtered and self.config.patch_names is not None:
            folder_path = self.storage.get_folder(self.config.patch_names.input_folder_key)
            patch_names_path = fs.path.join(folder_path, self.config.patch_names.filename)
            relevant_patch_names = set(load_names(self.storage.filesystem, patch_names_path))

            for crs, geoms in grid.items():
                grid[crs] = geoms[geoms[self.NAME_COLUMN].isin(relevant_patch_names)]

            grid = {crs: geoms for crs, geoms in grid.items() if not geoms.empty}

            num_geoms = sum(map(len, grid.values()))
            if len(relevant_patch_names) != num_geoms:
                raise ValueError(
                    f"Filtration done with {len(relevant_patch_names)} unique names, but {num_geoms} patches were found"
                )

        return grid

    @abstractmethod
    def _create_grid(self) -> Dict[CRS, gpd.GeoDataFrame]:
        """Defines a new grid, which encodes how the area is split into EOPatches.

        The grid is split into different CRS zones. The `bounds` properties of the geometries are taken as BBox
        definitions. EOPatch names are stored in a column with identifier `self.NAME_COLUMN`.
        """

    def _load_grid(self, grid_path: str) -> Dict[CRS, gpd.GeoDataFrame]:
        """A method that loads the bounding box grid from the cache folder."""
        LOGGER.info("Loading grid from %s", grid_path)

        grid = {}
        with LocalFile(grid_path, mode="r", filesystem=self.storage.filesystem) as local_file:
            for crs_layer in fiona.listlayers(local_file.path):
                data = gpd.read_file(local_file.path, layer=crs_layer, engine=self.storage.config.geopandas_backend)
                grid[CRS(data.crs)] = data

        return grid

    def _save_grid(self, grid: Dict[CRS, gpd.GeoDataFrame], grid_path: str) -> None:
        """A method that saves the bounding box grid to the cache folder."""
        LOGGER.info("Saving grid to %s", grid_path)

        with LocalFile(grid_path, mode="w", filesystem=self.storage.filesystem) as local_file:
            for _, crs_grid in grid.items():
                crs_grid.to_file(
                    local_file.path,
                    driver="GPKG",
                    encoding="utf-8",
                    layer=f"Grid EPSG:{crs_grid.crs.to_epsg()}",
                    engine=self.storage.config.geopandas_backend,
                )

    @abstractmethod
    def get_grid_cache_filename(self) -> str:
        """Provides a filename that is used for caching the grid, including the file extensions (likely .gpkg).

        Should ensure that two different grids don't clash.
        """

    def get_patch_list(self) -> PatchList:
        """Returns a list of eopatch names and appropriate BBoxes."""

        named_bboxes = []
        for crs, grid in self.get_grid(filtered=True).items():
            for _, row in grid.iterrows():
                named_bboxes.append((row.eopatch_name, BBox(row.geometry.bounds, crs=crs)))

        return named_bboxes


def get_geometry_from_file(
    filesystem: fs.base.FS, file_path: str, geopandas_engine: Literal["fiona", "pyogrio"] = "fiona"
) -> Geometry:
    """Provides a single geometry object of entire AOI"""
    with LocalFile(file_path, mode="r", filesystem=filesystem) as local_file:
        area_df = gpd.read_file(local_file.path, engine=geopandas_engine)
        area_shape = shapely.ops.unary_union(area_df.geometry)
        return Geometry(area_shape, CRS(area_df.crs))



"""
Module defining common validators for schemas and validator wrappers
"""
import datetime as dt
import inspect
from typing import TYPE_CHECKING, Any, Callable, Dict, Iterable, Optional, Tuple, Type, Union

import numpy as np
from pydantic import BaseModel, Field, validator

from eolearn.core import FeatureType
from eolearn.core.utils.parsing import parse_feature
from sentinelhub import DataCollection
from sentinelhub.data_collections_bands import Band, Bands, MetaBands, Unit

from ..types import Feature, RawSchemaDict, TimePeriod
from .meta import collect_schema, import_object

if TYPE_CHECKING:
    from ..core.schemas import ManagerSchema


def field_validator(field: str, validator_fun: Callable, allow_reuse: bool = True, **kwargs: Any) -> classmethod:
    """Sugared syntax for the `validator` decorator of `pydantic`"""
    return validator(field, allow_reuse=allow_reuse, **kwargs)(validator_fun)


def optional_field_validator(
    field: str, validator_fun: Callable, allow_reuse: bool = True, **kwargs: Any
) -> classmethod:
    """Wraps the validator functions so that `None` is always a valid input and only calls the validator on values.

    This allows re-use of validators e.g. if we have a validator for `Path` we can now use it for `Optional[Path]`.
    Because `pydantic` has a variable amount of arguments passed to the validator this function can only be used
    with validators that include `**kwargs` (or require all three arguments). For details on this behaviour
    consult the [validators documentation](https://pydantic-docs.helpmanual.io/usage/validators/).
    """
    # In order to propagate the pydantic python magic we need a bit of python magic ourselves
    additional_args = inspect.getfullargspec(validator_fun).args[1:]

    def optional_validator(value, values, config, field):  # type: ignore[no-untyped-def]
        if value is not None:
            all_kwargs = {"values": values, "config": config, "field": field}
            kwargs = {k: v for k, v in all_kwargs.items() if k in additional_args}
            return validator_fun(value, **kwargs)
        return None

    optional_validator.__name__ = f"optional_{validator_fun.__name__}"  # used for docbuilding purposes
    # the correct way would be to use `functools.wraps` but this breaks pydantics python magic

    return validator(field, allow_reuse=allow_reuse, **kwargs)(optional_validator)


def ensure_exactly_one_defined(first_param: str, second_param: str, **kwargs: Any) -> classmethod:
    """A validator (applied to `second_param` field) that makes sure only one of the two parameters is defined.

    Make sure that the definition of `second_param` comes after `first_param˙ (in line-order).
    """

    def ensure_exclusion(cls: type, value: Optional[Any], values: RawSchemaDict) -> Optional[Any]:
        is_param1_undefined = values.get(first_param) is None
        is_param2_undefined = value is None
        assert (
            is_param1_undefined != is_param2_undefined
        ), f"Exactly one of parameters `{first_param}` and `{second_param}` has to be specified."

        return value

    ensure_exclusion.__name__ = f"cannot_be_used_with_{first_param}"  # used for docbuilding purposes

    return field_validator(second_param, ensure_exclusion, **kwargs)


def ensure_defined_together(first_param: str, second_param: str, **kwargs: Any) -> classmethod:
    """A validator (applied to `second_param` field) that makes sure that the two parameters are both (un)defined.

    Make sure that the definition of `second_param` comes after `first_param˙ (in line-order).
    """

    def ensure_both(cls: type, value: Optional[Any], values: RawSchemaDict) -> Optional[Any]:
        is_param1_undefined = values.get(first_param) is None
        is_param2_undefined = value is None
        assert (
            is_param1_undefined == is_param2_undefined
        ), f"Both or neither of parameters `{first_param}` and `{second_param}` have to be specified."

        return value

    ensure_both.__name__ = f"must_be_used_with_{first_param}"  # used for docbuilding purposes

    return field_validator(second_param, ensure_both, **kwargs)


def ensure_storage_key_presence(key: str, **kwargs: Any) -> classmethod:
    """A field validator that makes sure that the specified storage key is present in the storage structure."""

    def validate_storage_key(cls: type, key: Optional[str], values: RawSchemaDict) -> Optional[str]:
        if key is not None:
            predefined_keys = ["input_data", "logs", "cache"]
            assert (
                key in values["storage"].structure or key in predefined_keys
            ), f"Couldn't find storage key {key!r} in the storage structure!"

        return key

    return field_validator(key, validate_storage_key, **kwargs)


def parse_time_period(value: Tuple[str, str]) -> TimePeriod:
    """Allows parsing of preset options of shape `[preset_kind, year]` but that requires `pre` validation"""
    presets = ["yearly", "season", "Q1", "Q2", "Q3", "Q4", "Q1-yearly", "Q2-yearly", "Q3-yearly", "Q4-yearly"]

    if value[0] in presets:
        kind, year = value[0], int(value[1])
        start_dates = {
            "yearly": f"{year}-01-01",
            "season": f"{year-1}-09-01",
            "Q1": f"{year}-01-01",
            "Q2": f"{year}-04-01",
            "Q3": f"{year}-07-01",
            "Q4": f"{year}-10-01",
            "Q1-yearly": f"{year-1}-04-01",
            "Q2-yearly": f"{year-1}-07-01",
            "Q3-yearly": f"{year-1}-10-01",
            "Q4-yearly": f"{year}-01-01",
        }
        end_dates = {
            "yearly": f"{year}-12-31",
            "season": f"{year}-09-01",
            "Q1": f"{year}-03-31",
            "Q2": f"{year}-06-30",
            "Q3": f"{year}-09-30",
            "Q4": f"{year}-12-31",
            "Q1-yearly": f"{year}-03-31",
            "Q2-yearly": f"{year}-06-30",
            "Q3-yearly": f"{year}-09-30",
            "Q4-yearly": f"{year}-12-31",
        }
        value = start_dates[kind], end_dates[kind]

    start = dt.datetime.strptime(value[0], "%Y-%m-%d").date()
    end = dt.datetime.strptime(value[1], "%Y-%m-%d").date()
    assert start <= end, "Invalid start and end dates provided. End date must follow the start date"
    return start, end


def parse_dtype(value: str) -> np.dtype:
    return np.dtype(value)


def validate_manager(value: dict) -> "ManagerSchema":
    """Parse and validate schema describing a manager."""
    assert "manager" in value, "Manager definition has no `manager` field that specifies its class."
    manager_class = import_object(value["manager"])
    manager_schema = collect_schema(manager_class)
    return manager_schema.parse_obj(value)  # type: ignore[return-value]


class BandSchema(BaseModel):
    """Schema used in parsing DataCollection bands."""

    name: str
    units: Tuple[Unit, ...]
    output_types: Tuple[type, ...]

    @validator("output_types", pre=True, each_item=True)
    def _parse_output_types(cls, value: str) -> type:
        if value == "bool":
            return bool
        return np.dtype(value).type


class DataCollectionSchema(BaseModel):
    """Schema used in parsing DataCollection objects. Extra parameters are passed to the definition as `**params`."""

    name: str = Field(
        "Name of the data collection. When defining BYOC collections use `BYOC_` prefix and for Batch collections use"
        " `BATCH_` to auto-generate fields with `define_byoc` or `define_batch`."
    )
    bands: Union[None, str, Tuple[BandSchema, ...]] = Field(
        None, description="Name of predefined collection in `Bands` or custom specification via `BandSchema`."
    )
    metabands: Union[None, str, Tuple[BandSchema, ...]] = Field(
        None, description="Name of predefined collection in `MetaBands` or custom specification via `BandSchema`."
    )

    class Config:
        extra = "allow"  # in order to pass on arbitrary parameters but keep definition shorter
        arbitrary_types_allowed = True


def _bands_parser(
    bands_collection: Union[Type[Bands], Type[MetaBands]], value: Union[None, str, Tuple[BandSchema, ...]]
) -> Optional[Tuple[Band, ...]]:
    """Collects defaults or parses bands from schemas."""

    if value is None:
        return None
    if isinstance(value, str):
        return getattr(bands_collection, value)
    return tuple(Band(band_schema.name, band_schema.units, band_schema.output_types) for band_schema in value)


def parse_data_collection(value: Union[str, dict, DataCollection]) -> DataCollection:
    """Validates and parses the data collection.

    If a string is given, then it tries to fetch a pre-defined collection. Otherwise it constructs a new collection
    according to the prefix of the name (`BYOC_` prefix to use `define_byoc` and `BATCH_` to use `define_batch`).
    """
    if isinstance(value, DataCollection):
        return value  # required in order to allow default values

    assert isinstance(value, (str, dict)), "Can only parse collection names or `DataCollectionSchema` definitions."

    params: Dict[str, Any] = {}
    if isinstance(value, str):
        name = value
        if name in DataCollection.__members__:
            return getattr(DataCollection, name)
    else:
        params = dict(DataCollectionSchema.parse_obj(value))
        params["bands"] = _bands_parser(Bands, params["bands"])
        params["metabands"] = _bands_parser(MetaBands, params["metabands"])
        name = params.pop("name")

    if name.startswith("BYOC_"):
        collection_id = name.split("_")[-1]
        return DataCollection.define_byoc(collection_id, **params)

    if name.startswith("BATCH_"):
        collection_id = name.split("_")[-1]
        return DataCollection.define_batch(collection_id, **params)

    return DataCollection.define(name, **params)


def restrict_types(
    allowed_feature_types: Union[Iterable[FeatureType], Callable[[FeatureType], bool]]
) -> Callable[[Feature], Feature]:
    """Validates a field representing a feature, where it restricts the possible feature types."""

    def validate_feature(value: Feature) -> Feature:
        parse_feature(feature=value, allowed_feature_types=allowed_feature_types)
        return value

    return validate_feature



"""
Utilities for filtering eopatch lists
"""
from concurrent.futures import ThreadPoolExecutor
from typing import Sequence

import fs
from fs.base import FS
from tqdm.auto import tqdm

from eolearn.core import FeatureType
from eolearn.core.eodata_io import FilesystemDataInfo, get_filesystem_data_info

from ..types import FeatureSpec, PatchList


def check_if_features_exist(
    filesystem: FS,
    eopatch_path: str,
    features: Sequence[FeatureSpec],
) -> bool:
    """Checks whether an EOPatch in the given location has all specified features saved"""
    try:
        existing_data = get_filesystem_data_info(filesystem, eopatch_path, features)
        meta_features = [spec for spec in features if isinstance(spec, FeatureType)]
        regular_features = [spec for spec in features if isinstance(spec, tuple)]

        if not all(_check_if_meta_feature_exists(ftype, existing_data) for ftype in meta_features):
            return False

        for ftype, fname in regular_features:
            if ftype == FeatureType.META_INFO:
                raise ValueError("Cannot check for a specific meta-info feature!")
            if ftype not in existing_data.features or fname not in existing_data.features[ftype]:
                return False
        return True

    except (IOError, fs.errors.ResourceNotFound):
        return False


def _check_if_meta_feature_exists(ftype: FeatureType, existing_data: FilesystemDataInfo) -> bool:
    if ftype == FeatureType.BBOX and existing_data.bbox is None:
        return False
    if ftype == FeatureType.TIMESTAMPS and existing_data.timestamps is None:
        return False
    if ftype == FeatureType.META_INFO and existing_data.meta_info is None:
        return False
    return True


def get_patches_with_missing_features(
    filesystem: FS,
    patches_folder: str,
    patch_list: PatchList,
    features: Sequence[FeatureSpec],
) -> PatchList:
    """Filters out names of those EOPatches that are missing some given features.

    :param filesystem: A filesystem object.
    :param patches_folder: A path to folder with EOPatches, relative to `filesystem` object.
    :param patch_list: A list of EOPatch names.
    :param features: A list of EOPatch features.
    :return: A sublist of `patch_list` with only EOPatch names that are missing some features.
    """
    eopatch_paths = [fs.path.combine(patches_folder, eopatch) for eopatch, _ in patch_list]

    def check_patch(eopatch_path: str) -> bool:
        return check_if_features_exist(filesystem, eopatch_path, features)

    with ThreadPoolExecutor() as executor:
        has_features_list = list(
            tqdm(
                executor.map(check_patch, eopatch_paths),
                total=len(eopatch_paths),
                desc="Checking EOPatches",
            )
        )

    return [eopatch for eopatch, has_features in zip(patch_list, has_features_list) if not has_features]



"""Tasks for spatial operations on EOPatches, used in grid-switching."""
from typing import Dict, List, Optional, Tuple, Union, cast

import numpy as np
from geopandas import GeoDataFrame

from eolearn.core import EOPatch, EOTask, FeatureType, deep_eq
from sentinelhub import CRS, BBox, bbox_to_resolution

from ..types import Feature, FeatureSpec
from ..utils.general import convert_to_int
from ..utils.vector import concat_gdf


class SpatialJoinTask(EOTask):
    """Spatially joins data from multiple EOPatches"""

    ERR = 1e-8

    def __init__(
        self,
        features: List[FeatureSpec],
        no_data_map: Dict[Feature, float],
        unique_columns_map: Dict[Feature, List[str]],
        raise_misaligned: bool = True,
    ):
        self.features = self.parse_features(features)
        self.no_data_map = no_data_map
        self.unique_columns_map = unique_columns_map
        self.raise_misaligned = raise_misaligned

    def _join_spatial_rasters(
        self, rasters: List[np.ndarray], bboxes: List[BBox], joined_bbox: BBox, no_data_value: float
    ) -> np.ndarray:
        """Joins all rasters into a single new rasters."""
        if len({raster.dtype for raster in rasters}) != 1:
            raise ValueError("Cannot join raster features with different dtypes")

        resolution = self._get_resolution(rasters, bboxes)
        joined_raster = self._get_joined_raster(joined_bbox, resolution, rasters[0], no_data_value)
        height, width = joined_raster.shape[-3:-1]

        for raster, bbox in zip(rasters, bboxes):
            slice_x, slice_y = get_array_slices(
                joined_bbox,
                bbox,
                resolution=resolution,
                raise_misaligned=self.raise_misaligned,
                limit_x=(0, width),
                limit_y=(0, height),
            )
            joined_raster[..., slice_y, slice_x, :] = raster

        return joined_raster

    def _get_resolution(self, rasters: List[np.ndarray], bboxes: List[BBox]) -> np.ndarray:
        """Checks that resolutions of given geospatial rasters don't differ more than a rounding error."""
        resolutions = []

        for raster, bbox in zip(rasters, bboxes):
            height, width = raster.shape[-3:-1]
            resolutions.append(np.array(bbox_to_resolution(bbox, width=width, height=height, meters=False)))

        mean_resolution = np.mean(resolutions, axis=0)
        for resolution in resolutions:
            diff = np.amax(np.absolute(resolution - mean_resolution))
            if diff > self.ERR:
                raise ValueError("Rasters have different spatial resolutions therefore they cannot be joined together")

        return mean_resolution

    def _get_joined_raster(
        self, bbox: BBox, resolution: np.ndarray, sample_raster: np.ndarray, no_data_value: float
    ) -> np.ndarray:
        """Provides an empty raster for a given bbox and resolution."""
        image_size = np.abs(np.array(bbox.upper_right) - np.array(bbox.lower_left)) / resolution
        try:
            width, height = convert_to_int(image_size, raise_diff=self.raise_misaligned)
        except ValueError as exception:
            raise ValueError(
                "Pixels from joined raster would be misaligned with pixels from original rasters"
            ) from exception

        time_dim = sample_raster.shape[:-3]  # Either empty or with a single element
        shape = (*time_dim, height, width, sample_raster.shape[-1])
        return np.full(shape, no_data_value, dtype=sample_raster.dtype)

    @staticmethod
    def _join_vector_data(dataframes: List[GeoDataFrame], unique_columns: Optional[List[str]]) -> GeoDataFrame:
        """Joins dataframes and optionally drops duplicated rows."""
        joined_dataframe = concat_gdf(dataframes)
        if unique_columns:
            return joined_dataframe.drop_duplicates(subset=unique_columns)
        return joined_dataframe

    def execute(self, *eopatches: EOPatch, bbox: BBox) -> EOPatch:
        """Spatially joins given EOPatches into a new EOPatch with given bounding box."""
        eopatches = tuple(eopatch for eopatch in eopatches if eopatch.get_features())
        for eopatch in eopatches:
            if eopatch.bbox is None:
                raise ValueError("All non-empty input EOPatches should have a bounding box")
            if eopatch.bbox.crs is not bbox.crs:
                raise ValueError("EOPatches must have the same CRS as the given bounding box")

        # Sorting EOPatches in unique order so that the values in the overlapping areas will always be computed in
        # the same way.
        eopatches = tuple(sorted(eopatches, key=lambda eop: repr(eop.bbox)))

        joined_eopatch = EOPatch(bbox=bbox)

        for feature in self.features:
            feature_type, _ = feature
            if feature_type is FeatureType.BBOX:
                continue

            data = [eopatch[feature] for eopatch in eopatches if feature in eopatch]
            if not data:
                continue

            if feature_type.is_spatial():
                feature = cast(Feature, feature)  # bbox and timestamp are discarded with above check
                if feature_type.is_array():
                    bboxes: List[BBox] = [patch.bbox for patch in eopatches if feature in patch]  # type: ignore[misc]
                    joined_data = self._join_spatial_rasters(data, bboxes, bbox, self.no_data_map[feature])
                else:
                    joined_data = self._join_vector_data(data, self.unique_columns_map.get(feature))
            else:
                joined_data = data[0]

                if not all(deep_eq(joined_data, item) for item in data[1:]):
                    raise ValueError(
                        f"Different EOPatches have different values of a non-spatial feature {feature}. It is not clear"
                        " how to join them."
                    )

            joined_eopatch[feature] = joined_data

        return joined_eopatch


class SpatialSliceTask(EOTask):
    """Spatially slices given EOPatch to create a new one."""

    def __init__(self, features: List[FeatureSpec], raise_misaligned: bool = True):
        self.features = self.parse_features(features)
        self.raise_misaligned = raise_misaligned

    def _slice_raster(self, raster: np.ndarray, raster_bbox: BBox, slice_bbox: BBox) -> np.ndarray:
        """Spatially slice a raster array."""
        height, width = raster.shape[-3:-1]
        slice_x, slice_y = get_array_slices(
            raster_bbox, slice_bbox, size=(width, height), raise_misaligned=self.raise_misaligned
        )
        return raster[..., slice_y, slice_x, :]

    @staticmethod
    def _filter_vector(gdf: GeoDataFrame, bbox: BBox) -> GeoDataFrame:
        """Spatially filters a GeoDataFrame."""
        bbox = bbox.transform(CRS(gdf.crs))
        intersects_bbox = gdf.geometry.intersects(bbox.geometry)
        return gdf[intersects_bbox].copy(deep=True)

    def execute(self, eopatch: EOPatch, *, bbox: BBox, skip: bool = False) -> EOPatch:
        """Spatially slices given EOPatch with given bounding box.

        Can be skipped in cases where the subbox won't be saved."""
        if skip:
            return eopatch

        main_bbox = eopatch.bbox
        if not main_bbox:
            raise ValueError("EOPatch is missing a bounding box")
        if main_bbox.crs is not bbox.crs:
            raise ValueError("Given bbox is not in the same CRS as EOPatch bbox")
        if not main_bbox.geometry.contains(bbox.geometry):
            raise ValueError("Given bbox must be fully contained in EOPatch's bbox")

        sliced_eopatch = EOPatch(bbox=bbox)

        for feature in self.features:
            feature_type, _ = feature
            if feature_type is FeatureType.BBOX:
                continue
            data = eopatch[feature]

            if feature_type.is_spatial():
                if feature_type.is_array():
                    sliced_data = self._slice_raster(data, main_bbox, bbox)
                else:
                    sliced_data = self._filter_vector(data, bbox)
            else:
                sliced_data = data

            sliced_eopatch[feature] = sliced_data

        return sliced_eopatch


def get_array_slices(
    bbox: BBox,
    slice_bbox: BBox,
    *,
    resolution: Optional[Union[np.ndarray, Tuple[float, float]]] = None,
    size: Optional[Union[np.ndarray, Tuple[int, int]]] = None,
    raise_misaligned: bool = True,
    limit_x: Optional[Tuple[int, int]] = None,
    limit_y: Optional[Tuple[int, int]] = None,
) -> Tuple[slice, slice]:
    """Slicing taken from eolearn.io.ImportFromTiffTask.

    :param bbox: A bounding box of initial array.
    :param slice_bbox: A bounding box of array to be sliced.
    :param resolution: A working resolution in CRS units.
    :param size: A working size.
    :param raise_misaligned: Whether to raise an error if the slice would be pixel misaligned the initial array.
    :param limit_x: If provided it will clip the horizontal slice to a given interval, should be used to clip
        slice_bbox to bbox.
    :param limit_y: If provided it will clip the vertical slice to a given interval, should be used to clip slice_bbox
        to bbox.
    :return: A slice over horizontal direction and a slice over vertical direction.
    """
    raster_upper_left = np.array([bbox.min_x, bbox.max_y])

    if size is not None and resolution is None:
        width, height = size
        raster_lower_right = np.array([bbox.max_x, bbox.min_y])
        resolution = np.abs((raster_upper_left - raster_lower_right)) / (width, height)
    elif size is None and resolution is not None:
        resolution = np.array(resolution)
    else:
        raise ValueError("Only one of the resolution and size can be given")

    slice_upper_left = np.array([slice_bbox.min_x, slice_bbox.max_y])
    slice_lower_right = np.array([slice_bbox.max_x, slice_bbox.min_y])
    axis_flip = [1, -1]  # image origin is upper left, geographic origin is lower left

    offset = axis_flip * (slice_upper_left - raster_upper_left) / resolution
    try:
        offset_x, offset_y = convert_to_int(offset, raise_diff=raise_misaligned)
        slice_size = abs(slice_lower_right - slice_upper_left) / resolution
        slice_width, slice_height = convert_to_int(slice_size, raise_diff=raise_misaligned)
    except ValueError as exception:
        raise ValueError(
            "Pixels from a slice raster would be misaligned with pixels from an original raster"
        ) from exception

    interval_x = offset_x, offset_x + slice_width
    if limit_x:
        interval_x = max(interval_x[0], limit_x[0]), min(interval_x[1], limit_x[1])
    interval_y = offset_y, offset_y + slice_height
    if limit_y:
        interval_y = max(interval_y[0], limit_y[0]), min(interval_y[1], limit_y[1])

    return slice(*interval_x), slice(*interval_y)



"""
Utilities for working with area grids
"""
from typing import List, Tuple

import numpy as np

from sentinelhub import BBox


def split_bbox(
    named_bbox: Tuple[str, BBox],
    split_x: int,
    split_y: int,
    buffer_x: float,
    buffer_y: float,
    naming_schema: str = "{name}_{i_x}_{i_y}",
) -> List[Tuple[str, BBox]]:
    """Splits a BBox into multiple smaller BBoxes with new names generated for them.

    The `buffer` parameters describe the buffer of the original BBox, which is copied to the split ones.
    """
    name, bbox = named_bbox
    min_x, min_y = bbox.lower_left
    max_x, max_y = bbox.upper_right

    x_edges, x_step = np.linspace(min_x + buffer_x, max_x - buffer_x, split_x, endpoint=False, retstep=True)
    y_edges, y_step = np.linspace(min_y + buffer_y, max_y - buffer_y, split_y, endpoint=False, retstep=True)

    split_bboxes = []
    for i_x, x_edge in enumerate(x_edges):
        for i_y, y_edge in enumerate(y_edges):
            lower_left = (x_edge - buffer_x, y_edge - buffer_y)
            upper_right = (x_edge + x_step + buffer_x, y_edge + y_step + buffer_y)
            split_bbox = BBox((lower_left, upper_right), crs=bbox.crs)

            split_name = naming_schema.format(name=name, i_x=i_x, i_y=i_y)
            split_bboxes.append((split_name, split_bbox))

    return split_bboxes



"""A collection of functions to be used with the ZipMapPipeline."""

from typing import Dict, Optional

import numpy as np

from ..core.schemas import BaseSchema
from .validators import optional_field_validator, parse_dtype


class MapParams(BaseSchema):
    mapping: Dict[int, int]
    default: Optional[int]
    dtype: np.dtype
    _parse_dtype = optional_field_validator("dtype", parse_dtype, pre=True)


def map_values(
    array: np.ndarray,
    *,
    mapping: Dict[int, int],
    default: Optional[int] = None,
    dtype: Optional[np.dtype] = None,
) -> np.ndarray:
    """Maps all values from `array` according to a dictionary. A default value can be given, which is assigned to
    values without a corresponding key in the mapping dictionary.

    Vectorizing the `.get` method is faster, but difficult to do with a default (and also loses speed advantage)."""

    if default is not None:
        flat_mapped_array = np.array([mapping.get(x, default) for x in array.ravel()], dtype=dtype)
        return flat_mapped_array.reshape(array.shape)

    mapped_array = np.vectorize(mapping.get)(array)
    return mapped_array.astype(dtype)



"""Implements a pipeline for rasterizing vector datasets."""
import logging
import os
import uuid
from typing import Any, Dict, List, Optional, Tuple, Union

import fiona
import fs
import geopandas as gpd
import numpy as np
from pydantic import Field, validator

from eolearn.core import CreateEOPatchTask, EONode, EOWorkflow, FeatureType, LoadTask, OverwritePermission, SaveTask
from eolearn.geometry import VectorToRasterTask
from eolearn.io import VectorImportTask
from sentinelhub import CRS

from ..core.pipeline import Pipeline
from ..core.schemas import BaseSchema
from ..types import Feature, FeatureSpec, PatchList
from ..utils.filter import get_patches_with_missing_features
from ..utils.fs import LocalFile
from ..utils.validators import ensure_exactly_one_defined, ensure_storage_key_presence, field_validator, parse_dtype
from ..utils.vector import concat_gdf

LOGGER = logging.getLogger(__name__)


class Preprocessing(BaseSchema):
    reproject_crs: Optional[int] = Field(
        description=(
            "An EPSG code of a CRS in which vector_input data will be reprojected once loaded. Mandatory if"
            " the input vector contains multiple layers in different CRS."
        ),
    )


class RasterizePipeline(Pipeline):
    """A pipeline module for rasterizing vector datasets."""

    class Schema(Pipeline.Schema):
        input_folder_key: str = Field(description="The storage manager key pointing to the input folder.")
        _ensure_input_folder_key = ensure_storage_key_presence("input_folder_key")
        output_folder_key: str = Field(description="The storage manager key pointing to the output folder.")
        _ensure_output_folder_key = ensure_storage_key_presence("output_folder_key")

        vector_input: Union[Feature, str] = Field(description="An input filename or a feature containing vector data.")
        output_feature: Feature = Field(description="A feature which should contain the newly rasterized data.")

        raster_value: Optional[float] = Field(
            description="Value to be used for all rasterized polygons. Cannot be used with `raster_values_column`."
        )
        raster_values_column: Optional[str] = Field(
            description=(
                "GeoPandas column for reading per-geometry rasterization values. Cannot be used with `raster_value`."
            )
        )
        _check_raster_value_values_column = ensure_exactly_one_defined("raster_value", "raster_values_column")

        resolution: Optional[float] = Field(
            description="Rendering resolution in meters. Cannot be used with `raster_shape`."
        )
        raster_shape: Optional[Tuple[int, int]] = Field(
            description="Shape of resulting raster image. Cannot be used with `resolution`."
        )
        _check_shape_resolution = ensure_exactly_one_defined("resolution", "raster_shape")

        dtype: np.dtype = Field(np.dtype("int32"), description="Numpy dtype of the output feature.")
        _parse_dtype = field_validator("dtype", parse_dtype, pre=True)

        preprocess_dataset: Optional[Preprocessing] = Field(
            description="Parameters used by `self.preprocess_dataset` method. Skipped if set to `None`."
        )
        dataset_layer: Optional[str] = Field(
            description="In case of multi-layer files, name of the layer with data to be rasterized."
        )
        polygon_buffer: float = Field(
            0, description="The size of polygon buffering to be applied before rasterization."
        )
        overlap_value: Optional[float] = Field(description="Value to write over the areas where polygons overlap.")
        no_data_value: int = Field(0, description="The no_data_value argument to be passed to VectorToRasterTask")
        compress_level: int = Field(1, description="Level of compression used in saving EOPatches")

        @validator("vector_input")
        def _check_vector_input(cls, vector_input: Union[Feature, str]) -> Union[Feature, str]:
            if isinstance(vector_input, str):
                assert vector_input.lower().endswith(
                    (".geojson", ".shp", ".gpkg", ".gdb")
                ), f"Input file path {vector_input} should be a GeoJSON, Shapefile, GeoPackage or GeoDataBase."
            else:
                assert vector_input[0].is_vector(), "Only vector-like input features are allowed!"
            return vector_input

        @validator("output_feature")
        def _check_temporal_nature_match(cls, output_feature: Feature, values: Dict[str, Any]) -> Feature:
            assert output_feature[0].is_image(), "Only image-like output features are allowed!"

            vector_input = values.get("vector_input")
            if vector_input and not isinstance(vector_input, str):
                assert (
                    vector_input[0].is_temporal() == output_feature[0].is_temporal()
                ), "Temporal natures of the vector input and the raster output must match!"
            return output_feature

    config: Schema

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)

        self.filename: Optional[str] = None
        self.vector_feature: Feature

        if not isinstance(self.config.vector_input, str):
            self.vector_feature = self.config.vector_input
        else:
            ftype = FeatureType.VECTOR if self._is_temporal(self.config.output_feature) else FeatureType.VECTOR_TIMELESS
            self.filename = self.config.vector_input
            self.vector_feature = (ftype, f"TEMP_{uuid.uuid4().hex}")

    def filter_patch_list(self, patch_list: PatchList) -> PatchList:
        filtered_patch_list = get_patches_with_missing_features(
            self.storage.filesystem,
            self.storage.get_folder(self.config.output_folder_key),
            patch_list,
            self._get_output_features(),
        )

        return filtered_patch_list

    def run_procedure(self) -> Tuple[List[str], List[str]]:
        if self.filename is not None and self.config.preprocess_dataset is not None:
            self.run_dataset_preprocessing(self.filename, self.config.preprocess_dataset)

        return super().run_procedure()

    def run_dataset_preprocessing(self, filename: str, preprocess_config: Preprocessing) -> None:
        """Loads datasets, applies preprocessing steps and saves them to a cache folder"""
        LOGGER.info("Preprocessing dataset %s", filename)
        gpd_engine = self.storage.config.geopandas_backend

        file_path = fs.path.combine(self.storage.get_input_data_folder(), filename)
        with LocalFile(file_path, mode="r", filesystem=self.storage.filesystem) as local_file:
            dataset_layers = [
                gpd.read_file(local_file.path, layer=layer, encoding="utf-8", engine=gpd_engine)
                for layer in fiona.listlayers(local_file.path)
            ]

        reproject_crs = None if preprocess_config.reproject_crs is None else CRS(preprocess_config.reproject_crs)
        dataset_gdf = concat_gdf(dataset_layers, reproject_crs=reproject_crs)
        dataset_gdf = self.preprocess_dataset(dataset_gdf)
        dataset_path = self._get_dataset_path(filename)

        with LocalFile(dataset_path, mode="w", filesystem=self.storage.filesystem) as local_file:
            dataset_gdf.to_file(local_file.path, encoding="utf-8", driver="GPKG", engine=gpd_engine)

    def build_workflow(self) -> EOWorkflow:
        """Creates workflow that is divided into the following sub-parts:

        1. loading data,
        2. preprocessing steps,
        3. rasterization of features,
        4. postprocessing steps,
        5. saving results
        """
        if self.filename is not None:
            create_node = EONode(CreateEOPatchTask())
            path = self._get_dataset_path(self.filename)
            import_task = VectorImportTask(
                feature=self.vector_feature,
                path=path,
                filesystem=self.storage.filesystem,
                layer=self.config.dataset_layer,
            )
            data_preparation_node = EONode(import_task, inputs=[create_node])
        else:
            features = [self.vector_feature, FeatureType.BBOX]
            if self._is_temporal(self.vector_feature):
                features.append(FeatureType.TIMESTAMPS)
            input_task = LoadTask(
                self.storage.get_folder(self.config.input_folder_key),
                filesystem=self.storage.filesystem,
                features=features,
            )
            data_preparation_node = EONode(input_task)

        preprocess_node = self.get_prerasterization_node(data_preparation_node)
        rasterization_node = self.get_rasterization_node(preprocess_node)
        postprocess_node = self.get_postrasterization_node(rasterization_node)

        save_task = SaveTask(
            self.storage.get_folder(self.config.output_folder_key),
            filesystem=self.storage.filesystem,
            features=self._get_output_features(),
            overwrite_permission=OverwritePermission.OVERWRITE_FEATURES,
            compress_level=self.config.compress_level,
        )
        save_node = EONode(save_task, inputs=[postprocess_node])

        return EOWorkflow.from_endnodes(save_node)

    def preprocess_dataset(self, dataframe: gpd.GeoDataFrame) -> gpd.GeoDataFrame:  # pylint: disable=no-self-use
        """Method for applying custom preprocessing steps on the entire dataset"""
        return dataframe

    def get_prerasterization_node(self, previous_node: EONode) -> EONode:  # pylint: disable=no-self-use
        """Builds node with tasks to be applied after loading vector feature but before rasterization"""
        return previous_node

    def get_rasterization_node(self, previous_node: EONode) -> EONode:
        """Builds nodes containing rasterization tasks"""

        return EONode(
            inputs=[previous_node],
            task=VectorToRasterTask(
                vector_input=self.vector_feature,
                raster_feature=self.config.output_feature,
                values_column=self.config.raster_values_column,
                values=self.config.raster_value,
                buffer=self.config.polygon_buffer,
                raster_resolution=self.config.resolution,
                raster_shape=self.config.raster_shape,
                raster_dtype=np.dtype(self.config.dtype),
                no_data_value=self.config.no_data_value,
                overlap_value=self.config.overlap_value,
            ),
        )

    def get_postrasterization_node(self, previous_node: EONode) -> EONode:  # pylint: disable=no-self-use
        """Builds node with tasks to be applied after rasterization"""
        return previous_node

    def _get_dataset_path(self, filename: str) -> str:
        """Provides a path from where dataset should be loaded into the workflow"""
        if self.config.preprocess_dataset is not None:
            folder = self.storage.get_cache_folder()
            filename = f"preprocessed_{filename}"
            filename = (os.path.splitext(filename))[0] + ".gpkg"
        else:
            folder = self.storage.get_input_data_folder()

        return fs.path.combine(folder, filename)

    def _get_output_features(self) -> List[FeatureSpec]:
        """Lists all features that are to be saved upon the pipeline completion"""
        features: List[FeatureSpec] = [self.config.output_feature, FeatureType.BBOX]
        if self._is_temporal(self.config.output_feature):
            features.append(FeatureType.TIMESTAMPS)

        return features

    @staticmethod
    def _is_temporal(feature: Feature) -> bool:
        f_type, _ = feature
        return f_type.is_temporal()



"""Area manager implementation for custom grids."""
import logging
from typing import Dict

import fs
import geopandas as gpd
import shapely.ops
from pydantic import Field

from sentinelhub import CRS, Geometry

from ...utils.vector import concat_gdf
from .base import BaseAreaManager

LOGGER = logging.getLogger(__name__)


class CustomGridAreaManager(BaseAreaManager):
    """Area manager that works with a pre-defined grid of EOPatches"""

    class Schema(BaseAreaManager.Schema):
        grid_folder_key: str = Field("input_data", description="Storage key pointing to the folder with the grid file.")
        grid_filename: str = Field(
            description=(
                "A Geopackage with a collection of bounding boxes and attributes that define EOPatches. If bounding"
                " boxes are in multiple CRS then each Geopackage layer should contain bounding boxes from one CRS."
            ),
            regex=r"^.+\..+$",
        )
        name_column: str = Field(description="Name of the column containing EOPatch names.")

    config: Schema

    def _create_grid(self) -> Dict[CRS, gpd.GeoDataFrame]:
        grid_path = fs.path.combine(self.storage.get_folder(self.config.grid_folder_key), self.config.grid_filename)
        grid = self._load_grid(grid_path)

        for crs, crs_gdf in grid.items():
            # Correct name of eopatch-name-column, drop all non-significant ones
            names = crs_gdf[self.config.name_column]
            grid[crs] = gpd.GeoDataFrame(geometry=crs_gdf.geometry, data={self.NAME_COLUMN: names}, crs=crs_gdf.crs)

        return grid

    def get_area_geometry(self, *, crs: CRS = CRS.WGS84) -> Geometry:
        all_grid_gdfs = list(self.get_grid(filtered=True).values())
        area_df = concat_gdf(all_grid_gdfs, reproject_crs=crs)

        LOGGER.info("Calculating a unary union of the area geometries")
        area_shape = shapely.ops.unary_union(area_df.geometry)
        return Geometry(area_shape, crs)

    def get_grid_cache_filename(self) -> str:
        input_filename = fs.path.basename(self.config.grid_filename)
        input_filename = input_filename.rsplit(".", 1)[0]

        return f"{self.__class__.__name__}_{input_filename}.gpkg"



"""Implementation of the base Pipeline class."""
import logging
import time
import uuid
from typing import Any, Dict, List, Optional, Tuple, Type, TypeVar

from eolearn.core import CreateEOPatchTask, EOExecutor, EONode, EOWorkflow, LoadTask, SaveTask, WorkflowResults
from eolearn.core.extra.ray import RayExecutor

from ..types import ExecKwargs, PatchList, ProcessingType
from ..utils.general import current_timestamp
from ..utils.meta import import_object
from ..utils.ray import handle_ray_connection
from .area.base import BaseAreaManager
from .base import EOGrowObject
from .config import RawConfig
from .logging import EOExecutionFilter, EOExecutionHandler, LoggingManager
from .schemas import ManagerSchema, PipelineSchema
from .storage import StorageManager

Self = TypeVar("Self", bound="Pipeline")

LOGGER = logging.getLogger(__name__)


class Pipeline(EOGrowObject):
    """A base class for execution of processing procedures which may or may not include running EOWorkflows, running
    EOExecutions, creating maps, etc.

    The functionalities of this class are:
        - collecting input arguments (either from command line or as an initialization parameter) and parsing them
        - preparing a list of patches
        - preparing execution arguments
        - running the pipeline, monitoring, and reporting
    """

    class Schema(PipelineSchema):
        """Configuration schema, describing input parameters and their types."""

    config: Schema

    def __init__(self, config: Schema, raw_config: Optional[RawConfig] = None):
        """
        :param config: A dictionary with configuration parameters
        :param raw_config: The configuration parameters pre-validation, for logging purposes only
        """
        super().__init__(config)
        self._raw_config = raw_config

        self.pipeline_id = self._new_pipeline_id()
        self.current_execution_name = "<Not executed yet>"

        self.storage: StorageManager = self._load_manager(config.storage)
        self.sh_config = self.storage.sh_config

        self.area_manager: BaseAreaManager = self._load_manager(config.area, storage=self.storage)
        self.logging_manager: LoggingManager = self._load_manager(config.logging, storage=self.storage)

    @property
    def _pipeline_name(self) -> str:
        return self.config.pipeline_name or self.__class__.__name__

    @classmethod
    def from_raw_config(cls: Type[Self], config: RawConfig, *args: Any, **kwargs: Any) -> Self:
        """Creates an object from a dictionary by constructing a validated config and use it to create the object."""
        validated_config = cls.Schema.parse_obj(config)
        if "raw_config" not in kwargs:
            kwargs["raw_config"] = config
        return cls(validated_config, *args, **kwargs)

    @staticmethod
    def _new_pipeline_id() -> str:
        """Provides a new random uuid of a pipeline"""
        return uuid.uuid4().hex[:10]

    @staticmethod
    def _load_manager(manager_config: ManagerSchema, **manager_params: Any) -> Any:
        """Loads a manager class and back-propagates parsed config

        :param manager_key: A config key name of a sub-config with manager parameters
        :param manager_params: Other parameters to initialize a manager class
        """
        if manager_config.manager is None:
            raise ValueError("Unable to load manager, field `manager` specifying it's class is missing.")
        manager_class = import_object(manager_config.manager)
        manager = manager_class(manager_config, **manager_params)
        return manager

    def get_pipeline_execution_name(self, pipeline_timestamp: str) -> str:
        """Returns the full name of the pipeline execution"""
        return f"{pipeline_timestamp}-{self._pipeline_name}-{self.pipeline_id}"

    def get_patch_list(self) -> PatchList:
        """Method which at the initialization prepares the list of EOPatches which will be used"""
        patch_list = self.area_manager.get_patch_list()

        if self.config.test_subset is not None:
            LOGGER.info("Filtering according to `test_subset` parameter.")
            indices = {x for x in self.config.test_subset if isinstance(x, int)}
            names = {x for x in self.config.test_subset if isinstance(x, str)}
            patch_list = [(name, bbox) for i, (name, bbox) in enumerate(patch_list) if (i in indices or name in names)]

            if len(patch_list) < len(self.config.test_subset):
                raise ValueError(
                    f"The parameter `test_subset` specifies {len(self.config.test_subset)} patches, but only"
                    f" {len(patch_list)} remain after filtration. Please recheck your input for `test_subset`."
                )

        if self.config.skip_existing:
            LOGGER.info("Checking which EOPatches can be skipped")
            filtered_patch_list = self.filter_patch_list(patch_list)

            skip_message = (
                "Skipped some EOPatches" if len(filtered_patch_list) < len(patch_list) else "No EOPatches were skipped"
            )
            LOGGER.info("%s, %d / %d remaining", skip_message, len(filtered_patch_list), len(patch_list))

            return filtered_patch_list

        return patch_list

    def filter_patch_list(self, patch_list: PatchList) -> PatchList:
        """Specifies which EOPatches should be skipped when `skip_existing` is enabled."""
        raise NotImplementedError("Method `filter_patch_list` must be implemented in order to use `skip_existing`")

    def get_execution_arguments(self, workflow: EOWorkflow, patch_list: PatchList) -> ExecKwargs:
        """Prepares execution arguments for each eopatch from a list of patches.

        The output should be a dictionary of form `{execution_name: {node: node_kwargs}}`. Execution names are usually
         names of EOPatches, but can be anything.

        :param workflow: A workflow for which arguments will be prepared
        """
        exec_kwargs = {}
        nodes = workflow.get_nodes()
        for name, bbox in patch_list:
            patch_args: Dict[EONode, Dict[str, Any]] = {}

            for node in nodes:
                if isinstance(node.task, (SaveTask, LoadTask)):
                    patch_args[node] = dict(eopatch_folder=name)

                if isinstance(node.task, CreateEOPatchTask):
                    patch_args[node] = dict(bbox=bbox)

            exec_kwargs[name] = patch_args
        return exec_kwargs

    def _init_processing(self) -> ProcessingType:
        """Figures out which execution mode is used and configures connection to Ray if required."""
        is_connected = handle_ray_connection(self.config.use_ray)
        if is_connected:
            return ProcessingType.RAY
        return ProcessingType.MULTI if self.config.workers > 1 else ProcessingType.SINGLE

    def run_execution(
        self,
        workflow: EOWorkflow,
        execution_kwargs: ExecKwargs,
        **executor_run_params: Any,
    ) -> Tuple[List[str], List[str], List[WorkflowResults]]:
        """A method which runs EOExecutor on given workflow with given execution parameters

        :param workflow: A workflow to be executed
        :param execution_kwargs: A dictionary mapping execution names to dictionaries holding execution arguments
        :param eopatch_list: A custom list of EOPatch names on which execution will run. If not specified, the default
            self.patch_list will be used
        :return: Lists of successfully/unsuccessfully executed EOPatch names and the result of the EOWorkflow execution
        """
        executor_class: Type[EOExecutor]

        execution_kind = self._init_processing()
        if execution_kind is ProcessingType.RAY:
            executor_class = RayExecutor
        else:
            executor_class = EOExecutor
            executor_run_params["workers"] = self.config.workers

        LOGGER.info("Starting %s for %d EOPatches", executor_class.__name__, len(execution_kwargs))

        # Unpacking manually to ensure order matches
        list_of_kwargs, execution_names = [], []
        for exec_name, exec_kwargs in execution_kwargs.items():
            list_of_kwargs.append(exec_kwargs)
            execution_names.append(exec_name)

        executor = executor_class(
            workflow,
            list_of_kwargs,
            execution_names=execution_names,
            save_logs=self.logging_manager.config.save_logs,
            logs_folder=self.logging_manager.get_pipeline_logs_folder(self.current_execution_name),
            filesystem=self.storage.filesystem,
            logs_filter=EOExecutionFilter(ignore_packages=self.logging_manager.config.eoexecution_ignore_packages),
            logs_handler_factory=EOExecutionHandler,
        )
        execution_results = executor.run(**executor_run_params)

        successful = [execution_names[idx] for idx in executor.get_successful_executions()]
        failed = [execution_names[idx] for idx in executor.get_failed_executions()]
        LOGGER.info(
            "%s finished with %d / %d success rate",
            executor_class.__name__,
            len(successful),
            len(successful) + len(failed),
        )

        if self.logging_manager.config.save_logs:
            executor.make_report(include_logs=self.logging_manager.config.include_logs_to_report)
            LOGGER.info("Saved EOExecution report to %s", executor.get_report_path(full_path=True))

        return successful, failed, execution_results

    def run(self) -> None:
        """The main method for pipeline execution. It sets up logging and runs the pipeline procedure."""
        timestamp = current_timestamp()
        self.current_execution_name = self.get_pipeline_execution_name(timestamp)

        handlers = self.logging_manager.start_logging(self.current_execution_name)
        try:
            self.logging_manager.update_pipeline_report(
                pipeline_execution_name=self.current_execution_name,
                pipeline_config=self.config,
                pipeline_raw_config=self._raw_config,
                pipeline_id=self.pipeline_id,
                pipeline_timestamp=timestamp,
            )

            LOGGER.info("Running %s", self._pipeline_name)

            pipeline_start = time.time()
            finished, failed = self.run_procedure()
            elapsed_time = time.time() - pipeline_start

            if failed:
                LOGGER.info(
                    "Pipeline finished with some errors! Check %s",
                    self.logging_manager.get_pipeline_logs_folder(self.current_execution_name, full_path=True),
                )
            else:
                LOGGER.info("Pipeline finished successfully!")

            self.logging_manager.update_pipeline_report(
                pipeline_execution_name=self.current_execution_name,
                pipeline_raw_config=self._raw_config,
                pipeline_config=self.config,
                pipeline_id=self.pipeline_id,
                pipeline_timestamp=timestamp,
                elapsed_time=elapsed_time,
            )

            self.logging_manager.save_eopatch_execution_status(
                pipeline_execution_name=self.current_execution_name, finished=finished, failed=failed
            )
        finally:
            self.logging_manager.stop_logging(handlers)

    def run_procedure(self) -> Tuple[List[str], List[str]]:
        """Execution procedure of pipeline. Can be overridden if needed.

        By default, builds the workflow by using a `build_workflow` method, which must be additionally implemented.

        :return: A pair of lists representing successful and unsuccessful executions.
        """
        if not hasattr(self, "build_workflow"):
            raise NotImplementedError(
                "Default implementation of the `run_procedure` method requires implementation of the `build_workflow`"
                " method."
            )
        workflow = self.build_workflow()
        patch_list = self.get_patch_list()
        exec_args = self.get_execution_arguments(workflow, patch_list)

        finished, failed, _ = self.run_execution(workflow, exec_args)
        return finished, failed



"""Common tasks shared between pipelines."""
from typing import Callable, List, Optional, Union

import numpy as np

from eolearn.core import EOPatch, EOTask, SaveTask
from eolearn.core.utils.parsing import parse_renamed_feature
from eolearn.geometry import MorphologicalOperations

from ..types import Feature


class ClassFilterTask(EOTask):
    """Run class specific morphological operation."""

    def __init__(
        self,
        feature: Feature,
        labels: List[int],
        morph_operation: Union[MorphologicalOperations, Callable],
        struct_elem: Optional[np.ndarray] = None,
    ):
        """Perform a morphological operation on a given feature mask

        :param feature: Feature to be modified
        :param labels: List of labels to be considered for morphological operation
        :param morph_operation: Type of morphological operation ot perform
        :param struct_elem: Structured element to be used. Taken from `ml_tools.MorphologicalStructFactory`
        """
        self.feature_name: Optional[str]
        self.new_feature_name: Optional[str]
        self.renamed_feature = parse_renamed_feature(feature)
        self.labels = labels

        if isinstance(morph_operation, MorphologicalOperations):
            self.morph_operation = MorphologicalOperations.get_operation(morph_operation)
        else:
            self.morph_operation = morph_operation
        self.struct_elem = struct_elem

    def execute(self, eopatch: EOPatch) -> EOPatch:
        feature_type, feature_name, new_feature_name = self.renamed_feature
        mask = eopatch[(feature_type, feature_name)].copy()

        for label in self.labels:
            label_mask = np.squeeze((mask == label), axis=-1)
            mask_mod = self.morph_operation(label_mask, self.struct_elem) * label
            mask_mod = mask_mod[..., np.newaxis]
            mask[mask == label] = mask_mod[mask == label]

        eopatch[(feature_type, new_feature_name)] = mask
        return eopatch


class SkippableSaveTask(SaveTask):
    """Same as `SaveTask` but can be skipped if the `eopatch_folder` is set to `None`."""

    def execute(self, eopatch: EOPatch, *, eopatch_folder: Optional[str] = "") -> EOPatch:
        if eopatch_folder is None:
            return eopatch
        return super().execute(eopatch, eopatch_folder=eopatch_folder)



"""Implements a pipeline to construct features for training/prediction."""
import logging
from typing import Dict, List, Optional, Tuple

import numpy as np
from pydantic import Field

from eolearn.core import (
    CopyTask,
    EONode,
    EOWorkflow,
    FeatureType,
    LoadTask,
    MergeFeatureTask,
    OverwritePermission,
    SaveTask,
)
from eolearn.features import LinearInterpolationTask, NormalizedDifferenceIndexTask, SimpleFilterTask
from eolearn.mask import JoinMasksTask

from ..core.pipeline import Pipeline
from ..core.schemas import BaseSchema
from ..tasks.features import (
    MaxNDVIMosaickingTask,
    MedianMosaickingTask,
    MosaickingTask,
    ValidDataFractionPredicate,
    join_valid_and_cloud_masks,
)
from ..types import Feature, FeatureSpec, PatchList, TimePeriod
from ..utils.filter import get_patches_with_missing_features
from ..utils.validators import (
    ensure_storage_key_presence,
    field_validator,
    optional_field_validator,
    parse_dtype,
    parse_time_period,
)

LOGGER = logging.getLogger(__name__)


class ValidityFiltering(BaseSchema):
    cloud_mask_feature_name: Optional[str] = Field(
        description="Name of cloud mask to enable additional filtering by cloud"
    )
    valid_data_feature_name: str = Field(description="Name of the valid-data mask to use for filtering.")

    validity_threshold: Optional[float] = Field(
        description="Threshold to remove frames with valid data lower than threshold"
    )


class FeaturesPipeline(Pipeline):
    """A pipeline to calculate and prepare features for ML"""

    class Schema(Pipeline.Schema):
        input_folder_key: str = Field(
            description="The storage manager key pointing to the input folder for the features pipeline."
        )
        _ensure_input_folder_key = ensure_storage_key_presence("input_folder_key")
        output_folder_key: str = Field(
            description="The storage manager key pointing to the output folder for the features pipeline."
        )
        _ensure_output_folder_key = ensure_storage_key_presence("output_folder_key")

        bands_feature_name: str = Field(description="Name of data feature containing band data")

        data_preparation: ValidityFiltering

        ndis: Dict[str, Tuple[int, int]] = Field(
            default_factory=dict,
            description=(
                "A dictionary of kind `{feature_name: (id1, id2)}` that specifies how to calculate the NDIs of bands "
                "(with indices `id1` and `id2` in the bands feature) and save it under `feature_name`."
            ),
        )

        dtype: Optional[np.dtype] = Field(description="The dtype under which the concatenated features should be saved")
        _parse_dtype = optional_field_validator("dtype", parse_dtype, pre=True)
        output_feature_name: str = Field(description="Name of output data feature encompassing bands and NDIs")
        compress_level: int = Field(1, description="Level of compression used in saving eopatches")

    config: Schema

    def filter_patch_list(self, patch_list: PatchList) -> PatchList:
        """EOPatches are filtered according to existence of specified output features"""

        filtered_patch_list = get_patches_with_missing_features(
            self.storage.filesystem,
            self.storage.get_folder(self.config.output_folder_key),
            patch_list,
            self._get_output_features(),
        )

        return filtered_patch_list

    def _get_output_features(self) -> List[FeatureSpec]:
        """Lists all features that are to be saved upon the pipeline completion"""
        return [(FeatureType.DATA, self.config.output_feature_name), FeatureType.BBOX, FeatureType.TIMESTAMPS]

    def _get_bands_feature(self) -> Feature:
        return FeatureType.DATA, self.config.bands_feature_name

    def _get_valid_data_feature(self) -> Feature:
        return FeatureType.MASK, self.config.data_preparation.valid_data_feature_name

    def build_workflow(self) -> EOWorkflow:
        """
        Creates a workflow:
        1. Loads and prepares a 'bands_feature' and 'valid_data_feature'
        2. Temporally regularizes bands and NDIs
        3. Calculates NDIs based on 'bands_feature'
        4. Applies post-processing, which prepares all output features
        5. Saves all relevant features (specified in _get_output_features)
        """
        preparation_node = self.get_data_preparation_node()
        regularization_node = self.get_temporal_regularization_node(preparation_node)
        ndi_node = self.get_ndi_node(regularization_node)
        postprocessing_node = self.get_postprocessing_node(ndi_node)

        save_task = SaveTask(
            self.storage.get_folder(self.config.output_folder_key),
            filesystem=self.storage.filesystem,
            features=self._get_output_features(),
            compress_level=self.config.compress_level,
            overwrite_permission=OverwritePermission.OVERWRITE_FEATURES,
        )
        save_node = EONode(save_task, inputs=[postprocessing_node])

        return EOWorkflow.from_endnodes(save_node)

    def get_data_preparation_node(self) -> EONode:
        """Nodes that load, filter, and prepare a feature containing all bands

        :return: A node with preparation tasks and feature for masking invalid data
        """
        filtering_config = self.config.data_preparation

        load_task = LoadTask(
            self.storage.get_folder(self.config.input_folder_key),
            filesystem=self.storage.filesystem,
            lazy_loading=True,
        )
        end_node = EONode(load_task)

        if filtering_config.cloud_mask_feature_name:
            zip_masks_task = JoinMasksTask(
                [
                    self._get_valid_data_feature(),
                    (FeatureType.MASK, filtering_config.cloud_mask_feature_name),
                ],
                self._get_valid_data_feature(),
                join_operation=join_valid_and_cloud_masks,
            )
            end_node = EONode(zip_masks_task, inputs=[end_node], name="Combine validity mask and cloud mask")

        if filtering_config.validity_threshold is not None:
            filter_func = ValidDataFractionPredicate(filtering_config.validity_threshold)
            filter_task = SimpleFilterTask(self._get_valid_data_feature(), filter_func)
            end_node = EONode(filter_task, inputs=[end_node])

        return end_node

    def get_temporal_regularization_node(self, previous_node: EONode) -> EONode:
        """Builds node adding temporal regularization to workflow."""
        return previous_node

    def get_ndi_node(self, previous_node: EONode) -> EONode:
        """Builds a node for constructing Normalized Difference Indices"""

        for name, (id1, id2) in self.config.ndis.items():
            ndi_task = NormalizedDifferenceIndexTask(self._get_bands_feature(), (FeatureType.DATA, name), [id1, id2])
            previous_node = EONode(ndi_task, inputs=[previous_node])

        return previous_node

    def get_postprocessing_node(self, previous_node: EONode) -> EONode:
        """Tasks performed after temporal regularization. Should also prepare features for the saving step"""
        ndi_features = [(FeatureType.DATA, name) for name in self.config.ndis]
        merge_task = MergeFeatureTask(
            [self._get_bands_feature(), *ndi_features],
            (FeatureType.DATA, self.config.output_feature_name),
            dtype=self.config.dtype,
        )
        return EONode(merge_task, inputs=[previous_node])


class InterpolationSpecifications(BaseSchema):
    time_period: TimePeriod
    _parse_time_period = field_validator("time_period", parse_time_period, pre=True)
    resampling_period: int


class InterpolationFeaturesPipeline(FeaturesPipeline):
    """A pipeline to calculate and prepare features for ML including interpolation"""

    class Schema(FeaturesPipeline.Schema):
        interpolation: Optional[InterpolationSpecifications] = Field(
            description=(
                "Fine-tuning of interpolation parameters. If not set, the interpolation will work on current timestamps"
            )
        )

    config: Schema

    def get_temporal_regularization_node(self, previous_node: EONode) -> EONode:
        resample_range = None
        if self.config.interpolation:
            start, end = self.config.interpolation.time_period
            start_time, end_time = start.strftime("%Y-%m-%d"), end.strftime("%Y-%m-%d")
            resample_range = (start_time, end_time, self.config.interpolation.resampling_period)

        interpolation_task = LinearInterpolationTask(
            feature=self._get_bands_feature(),
            mask_feature=self._get_valid_data_feature(),
            resample_range=resample_range,
            bounds_error=False,
        )
        return EONode(interpolation_task, inputs=[previous_node])


class MosaickingSpecifications(BaseSchema):
    time_period: TimePeriod
    _parse_time_period = field_validator("time_period", parse_time_period, pre=True)
    n_mosaics: int

    max_ndi_indices: Optional[Tuple[int, int]] = Field(
        description=(
            "When omitted uses median value mosaicking. If set, uses max NDI mosaicking for the NDI of the bands at"
            " specified indices. For example, to use max NDVI when using all 13 bands of L1C set parameter to `[7, 3]`"
            " (uses B08 and B04)"
        )
    )


class MosaickingFeaturesPipeline(FeaturesPipeline):
    """A pipeline to calculate and prepare features for ML including mosaicking"""

    _NDI_FEATURE = FeatureType.DATA, "_NDI_FEATURE_OF_MOSAICKING_PIPELINE"

    class Schema(FeaturesPipeline.Schema):
        mosaicking: MosaickingSpecifications = Field(
            description=(
                "Fine-tuning of mosaicking parameters. If not set, the interpolation will work on current timestamps"
            )
        )

    config: Schema

    def get_data_preparation_node(self) -> EONode:
        preparation_node = super().get_data_preparation_node()

        if self.config.mosaicking.max_ndi_indices:
            ndi_task = NormalizedDifferenceIndexTask(
                self._get_bands_feature(), self._NDI_FEATURE, self.config.mosaicking.max_ndi_indices
            )
            return EONode(ndi_task, inputs=[preparation_node])

        return preparation_node

    def get_temporal_regularization_node(self, previous_node: EONode) -> EONode:
        start_date, end_date = self.config.mosaicking.time_period

        mosaicking_task: MosaickingTask
        if self.config.mosaicking.max_ndi_indices:
            mosaicking_task = MaxNDVIMosaickingTask(
                self._get_bands_feature(),
                (start_date, end_date, self.config.mosaicking.n_mosaics),
                self._NDI_FEATURE,
                self._get_valid_data_feature(),
            )
        else:
            mosaicking_task = MedianMosaickingTask(
                self._get_bands_feature(),
                (start_date, end_date, self.config.mosaicking.n_mosaics),
                self._get_valid_data_feature(),
            )
        mosaicking_node = EONode(mosaicking_task, inputs=[previous_node])
        return EONode(
            CopyTask(features=[self._get_bands_feature(), FeatureType.BBOX, FeatureType.TIMESTAMPS]),
            inputs=[mosaicking_node],
            name="Remove non-mosaicked features",
        )



"""Implements a pipeline for importing vector data from a file."""
import fs
from pydantic import Field

from eolearn.core import EONode, EOWorkflow, FeatureType, OverwritePermission, SaveTask
from eolearn.io import VectorImportTask

from ..core.pipeline import Pipeline
from ..types import ExecKwargs, Feature, PatchList
from ..utils.validators import ensure_storage_key_presence, field_validator, restrict_types


class ImportVectorPipeline(Pipeline):
    class Schema(Pipeline.Schema):
        input_folder_key: str = Field("input_data", description="The folder key into which the EOPatch will be saved. ")
        _ensure_input_folder_key = ensure_storage_key_presence("input_folder_key")
        input_filename: str = Field(
            description="Filename of the vector file to be imported. Needs to be located in the input-data folder."
        )
        reproject: bool = Field(True, description="Controls whether the vector file is reprojected to the EOPatch CRS.")
        clip: bool = Field(True, description="Controls whether the polygons are clipped to the EOPatch boundaries. ")
        output_feature: Feature = Field(description="The EOPatch feature to which the vector will be imported.")
        output_folder_key: str = Field(description="The folder key into which the EOPatch will be saved. ")
        _ensure_output_folder_key = ensure_storage_key_presence("output_folder_key")

        _restrict_output_feature = field_validator(
            "output_feature", restrict_types([FeatureType.VECTOR, FeatureType.VECTOR_TIMELESS])
        )

    config: Schema

    def get_execution_arguments(self, workflow: EOWorkflow, patch_list: PatchList) -> ExecKwargs:
        exec_kwargs = super().get_execution_arguments(workflow, patch_list)
        vector_node = next(node for node in workflow.get_nodes() if isinstance(node.task, VectorImportTask))

        for name, bbox in patch_list:
            exec_kwargs[name][vector_node] = dict(bbox=bbox)

        return exec_kwargs

    def build_workflow(self) -> EOWorkflow:
        path = fs.path.join(self.storage.get_input_data_folder(), self.config.input_filename)

        vector_import_task = VectorImportTask(
            feature=self.config.output_feature,
            path=path,
            reproject=self.config.reproject,
            clip=self.config.clip,
            filesystem=self.storage.filesystem,
        )
        import_node = EONode(vector_import_task)

        save_task = SaveTask(
            path=self.storage.get_folder(self.config.output_folder_key),
            filesystem=self.storage.filesystem,
            features=[FeatureType.BBOX, self.config.output_feature],
            compress_level=1,
            overwrite_permission=OverwritePermission.OVERWRITE_FEATURES,
        )
        save_node = EONode(save_task, inputs=[import_node])
        return EOWorkflow.from_endnodes(save_node)



"""Implements a base prediction pipeline and a LGBM specialized classification and regression pipelines."""
import abc
from typing import List, Optional, Tuple

import fs
import numpy as np
from pydantic import Field

from eolearn.core import EONode, EOWorkflow, FeatureType, LoadTask, MergeEOPatchesTask, OverwritePermission, SaveTask

from ..core.pipeline import Pipeline
from ..tasks.prediction import ClassificationPredictionTask, RegressionPredictionTask
from ..types import Feature, FeatureSpec, PatchList
from ..utils.filter import get_patches_with_missing_features
from ..utils.validators import (
    ensure_defined_together,
    ensure_storage_key_presence,
    optional_field_validator,
    parse_dtype,
)


class BasePredictionPipeline(Pipeline, metaclass=abc.ABCMeta):
    """Pipeline to load a model and run prediction on EOPatches"""

    class Schema(Pipeline.Schema):
        input_folder_key: str = Field(
            description="The storage manager key pointing to the input folder of the model input data."
        )
        _ensure_input_folder_key = ensure_storage_key_presence("input_folder_key")
        input_features: List[Feature] = Field(
            description=(
                "List of features of form `[(feature_type, feature_name)]` specifying which features are model input in"
                " the correct order"
            )
        )

        output_folder_key: str = Field(
            description="The storage manager key pointing to the output folder for the prediction pipeline."
        )
        _ensure_output_folder_key = ensure_storage_key_presence("output_folder_key")

        dtype: Optional[np.dtype] = Field(
            description="Casts the result to desired type. Uses predictor output type by default."
        )
        _parse_dtype = optional_field_validator("dtype", parse_dtype, pre=True)

        prediction_mask_feature_name: Optional[str] = Field(
            description="Name of `MASK_TIMELESS` feature which defines which areas will be predicted"
        )
        prediction_mask_folder_key: Optional[str]
        _ensure_mask_feature_key = ensure_defined_together("prediction_mask_feature_name", "prediction_mask_folder_key")

        model_folder_key: str = Field(
            description="The storage manager key pointing to the folder of the model used in the prediction pipeline."
        )
        _ensure_model_folder_key = ensure_storage_key_presence("model_folder_key")
        compress_level: int = Field(1, description="Level of compression used in saving EOPatches")

    config: Schema

    @abc.abstractmethod
    def _get_output_features(self) -> List[FeatureSpec]:
        """Lists all features that are to be saved upon the pipeline completion"""

    @property
    def _is_mp_lock_needed(self) -> bool:
        """If a multiprocessing lock is needed when executing"""
        return not self.config.use_ray and self.config.workers > 1

    def filter_patch_list(self, patch_list: PatchList) -> PatchList:
        """EOPatches are filtered according to existence of specified output features"""

        filtered_patch_list = get_patches_with_missing_features(
            self.storage.filesystem,
            self.storage.get_folder(self.config.output_folder_key),
            patch_list,
            self._get_output_features(),
        )

        return filtered_patch_list

    def build_workflow(self) -> EOWorkflow:
        """Workflow handling the prediction for eopatches.
        The workflow allows to add smoothing and custom thresholding to the predicted scores
        """
        preparation_node = self._get_data_preparation_node()
        predictions_node = self._get_prediction_node(preparation_node)
        saving_node = self._get_saving_node(predictions_node)

        return EOWorkflow.from_endnodes(saving_node)

    def _get_data_preparation_node(self) -> EONode:
        """Returns nodes containing for loading and preparing the data as well as the endpoint tasks"""
        features_load_node = EONode(
            LoadTask(
                self.storage.get_folder(self.config.input_folder_key),
                filesystem=self.storage.filesystem,
                features=[FeatureType.BBOX, FeatureType.TIMESTAMPS, *self.config.input_features],
            )
        )

        if not self.config.prediction_mask_folder_key:
            return features_load_node

        mask_load_node = EONode(
            LoadTask(
                self.storage.get_folder(self.config.prediction_mask_folder_key),
                filesystem=self.storage.filesystem,
                features=[(FeatureType.MASK_TIMELESS, self.config.prediction_mask_feature_name)],
            )
        )

        return EONode(MergeEOPatchesTask(), inputs=[features_load_node, mask_load_node])

    @abc.abstractmethod
    def _get_prediction_node(self, previous_node: EONode) -> EONode:
        """Returns nodes for using the model on the prepared data and the endpoints of the dependencies"""

    def _get_saving_node(self, previous_node: EONode) -> EONode:
        """Returns nodes for finalizing and saving features"""
        save_task = SaveTask(
            self.storage.get_folder(self.config.output_folder_key),
            filesystem=self.storage.filesystem,
            features=self._get_output_features(),
            overwrite_permission=OverwritePermission.OVERWRITE_FEATURES,
            compress_level=self.config.compress_level,
        )

        return EONode(save_task, inputs=[previous_node])


class RegressionPredictionPipeline(BasePredictionPipeline):
    class Schema(BasePredictionPipeline.Schema):
        output_feature_name: str
        model_filename: str = Field(description="A filename of a regression model to be used for prediction.")
        clip_predictions: Optional[Tuple[float, float]] = Field(
            description="Whether to clip values of predictions to specified interval"
        )

    config: Schema

    def _get_output_features(self) -> List[FeatureSpec]:
        return [FeatureType.BBOX, (FeatureType.DATA_TIMELESS, self.config.output_feature_name)]

    def _get_prediction_node(self, previous_node: EONode) -> EONode:
        model_path = fs.path.join(self.storage.get_folder(self.config.model_folder_key), self.config.model_filename)
        prediction_task = RegressionPredictionTask(
            model_path=model_path,
            filesystem=self.storage.filesystem,
            input_features=self.config.input_features,
            mask_feature=_optional_typed_feature(FeatureType.MASK_TIMELESS, self.config.prediction_mask_feature_name),
            output_feature=(FeatureType.DATA_TIMELESS, self.config.output_feature_name),
            output_dtype=self.config.dtype,
            mp_lock=self._is_mp_lock_needed,
            clip_predictions=self.config.clip_predictions,
        )
        return EONode(prediction_task, inputs=[previous_node])


class ClassificationPredictionPipeline(BasePredictionPipeline):
    class Schema(BasePredictionPipeline.Schema):
        output_feature_name: str
        output_probability_feature_name: Optional[str]

        model_filename: str = Field(description="A filename of a classification model to be used for prediction.")
        label_encoder_filename: Optional[str] = Field(
            description=(
                "Whether the predictions need to be decoded. The label encoder should be in the same model folder."
            )
        )

    config: Schema

    def _get_output_features(self) -> List[FeatureSpec]:
        features: List[FeatureSpec] = [FeatureType.BBOX, (FeatureType.MASK_TIMELESS, self.config.output_feature_name)]
        if self.config.output_probability_feature_name:
            features.append((FeatureType.DATA_TIMELESS, self.config.output_probability_feature_name))
        return features

    def _get_prediction_node(self, previous_node: EONode) -> EONode:
        model_path = fs.path.join(self.storage.get_folder(self.config.model_folder_key), self.config.model_filename)
        prediction_task = ClassificationPredictionTask(
            model_path=model_path,
            filesystem=self.storage.filesystem,
            input_features=self.config.input_features,
            mask_feature=_optional_typed_feature(FeatureType.MASK_TIMELESS, self.config.prediction_mask_feature_name),
            output_feature=(FeatureType.MASK_TIMELESS, self.config.output_feature_name),
            output_probability_feature=_optional_typed_feature(
                FeatureType.DATA_TIMELESS, self.config.output_probability_feature_name
            ),
            output_dtype=self.config.dtype,
            mp_lock=self._is_mp_lock_needed,
            label_encoder_filename=self.config.label_encoder_filename,
        )
        return EONode(prediction_task, inputs=[previous_node])


def _optional_typed_feature(ftype: FeatureType, fname: Optional[str]) -> Optional[Feature]:
    """Constructs a typed feature if possible"""
    return (ftype, fname) if fname is not None else None



"""Implements the command line interface for `eo-grow`."""
import json
import os
import re
import subprocess
from tempfile import NamedTemporaryFile
from typing import Optional, Tuple

import click

from .core.config import collect_configs_from_path, interpret_config_from_dict
from .core.schemas import build_schema_template
from .pipelines.testing import TestPipeline
from .utils.general import jsonify
from .utils.meta import collect_schema, import_object, load_pipeline_class
from .utils.ray import generate_cluster_config_path, start_cluster_if_needed

variables_option = click.option(
    "-v",
    "--variable",
    "cli_variables",
    multiple=True,
    type=str,
    help='Specify variables to use in config (overriding present config values). Must be of form `"name:value"`',
)
test_patches_option = click.option(
    "-t",
    "--test",
    "test_patches",
    multiple=True,
    type=int,
    help=(
        "One or more indices of EOPatches. If given, a pipeline will run only for those EOPatches. "
        "Example: -t 0 -t 42 will run the pipeline for EOPatches with indices 0 and 42."
    ),
)


@click.command()
@click.argument("config_path", type=click.Path())
@variables_option
@test_patches_option
def run_pipeline(config_path: str, cli_variables: Tuple[str, ...], test_patches: Tuple[int, ...]) -> None:
    """Execute eo-grow pipeline using CLI.

    \b
    Example:
        eogrow config_files/config.json
    """

    raw_configs = collect_configs_from_path(config_path)
    cli_variable_mapping = dict(_parse_cli_variable(cli_var) for cli_var in cli_variables)

    pipelines = []
    for raw_config in raw_configs:
        config = interpret_config_from_dict(raw_config, cli_variable_mapping)
        if test_patches:
            config["test_subset"] = list(test_patches)

        pipelines.append(load_pipeline_class(config).from_raw_config(config))

    for pipeline in pipelines:
        pipeline.run()


@click.command()
@click.argument("cluster_yaml", type=click.Path())
@click.argument("config_path", type=click.Path())
@click.option(
    "--start", "start_cluster", is_flag=True, type=bool, help="Starts the cluster if it is not currently running."
)
@click.option(
    "--stop",
    "stop_cluster",
    is_flag=True,
    type=bool,
    help=(
        "Stops the cluster if after running the pipeline. In order for this to work got to AWS console "
        "-> IAM -> Roles -> select ray-autoscaler-v1 role and attach IAMReadOnlyAccess policy."
    ),
)
@click.option(
    "--screen",
    "use_screen",
    is_flag=True,
    type=bool,
    help=(
        "Run the cluster in a detached mode using screen software. Use Ctrl+A+D to detach any time, "
        "even when running a pipeline or a Jupyter notebook. Use Ctrl+D to terminate the remote screen."
    ),
)
@click.option(
    "--tmux",
    "use_tmux",
    is_flag=True,
    type=bool,
    help="Run the cluster in a detached mode using tmux software. Use Ctrl+B and d to detach.",
)
@variables_option
@test_patches_option
def run_pipeline_on_cluster(
    config_path: str,
    cluster_yaml: str,
    start_cluster: bool,
    stop_cluster: bool,
    use_screen: bool,
    use_tmux: bool,
    cli_variables: Tuple[str, ...],
    test_patches: Tuple[int, ...],
) -> None:
    """Command for running an eo-grow pipeline on a remote Ray cluster of AWS EC2 instances. The provided config is
    fully constructed and uploaded to the cluster head in the `~/.synced_configs/` directory, where it is then
    executed. A custom suffix is added to distinguish runs which use the same config multiple times.

    \b
    Example:
        eogrow-ray cluster.yaml config_files/config.json
    """
    if start_cluster:
        start_cluster_if_needed(cluster_yaml)

    if stop_cluster and (use_screen or use_tmux):
        raise NotImplementedError("It is not clear how to combine stop flag with either screen or tmux flag")

    raw_configs = [interpret_config_from_dict(config) for config in collect_configs_from_path(config_path)]
    remote_path = generate_cluster_config_path(config_path)

    with NamedTemporaryFile(mode="w", delete=True, suffix=".json") as local_path:
        json.dump(raw_configs, local_path)
        subprocess.run(f"ray rsync_up {cluster_yaml} {local_path.name!r} {remote_path!r}", shell=True)

    cmd = (
        f"eogrow {remote_path}"
        + "".join(f' -v "{cli_var_spec}"' for cli_var_spec in cli_variables)  # noqa B028
        + "".join(f" -t {patch_index}" for patch_index in test_patches)
        + ("; " if stop_cluster else "")  # Otherwise, ray will incorrectly prepare a command for stopping a cluster
    )
    flag_info = [("stop", stop_cluster), ("screen", use_screen), ("tmux", use_tmux)]
    exec_flags = " ".join(f"--{flag_name}" for flag_name, use_flag in flag_info if use_flag)

    subprocess.run(f"ray exec {exec_flags} {cluster_yaml} {cmd!r}", shell=True)  # noqa B028


@click.command()
@click.argument("import_path", type=str)
@click.argument("template_path", type=click.Path(), required=False)
@click.option(
    "-f",
    "--force",
    "force_override",
    is_flag=True,
    type=bool,
    help=(
        "In case a template path is provided and a file in the path already exists this flag is used to force "
        "override it."
    ),
)
@click.option(
    "--template-format",
    "template_format",
    type=click.Choice(["minimal", "open-api"], case_sensitive=False),
    help="Specifies which template format to use. The default is `minimal`",
    default="minimal",
)
@click.option(
    "--required-only",
    "required_only",
    is_flag=True,
    type=bool,
    help="If provided it will only include required fields in the template. Only for `minimal` template format",
)
@click.option(
    "--add-descriptions",
    "add_descriptions",
    is_flag=True,
    type=bool,
    help="Adds descriptions to template. Only for `minimal` template format",
)
def make_template(
    import_path: str,
    template_path: Optional[str],
    force_override: bool,
    template_format: str,
    required_only: bool,
    add_descriptions: bool,
) -> None:
    """Command for creating a config template for an eo-grow pipeline

    \b
    Examples:
        - save template to file:
            eogrow-template eogrow.pipelines.download.DownloadPipeline config_files/download.json
        - print template to command line:
            eogrow-template eogrow.pipelines.download.DownloadPipeline
    """
    if not force_override and template_path and os.path.isfile(template_path):
        raise FileExistsError(f"File {template_path} already exists. You can use -f to force override it.")

    class_with_schema = import_object(import_path)
    schema = collect_schema(class_with_schema)

    if template_format == "open-api":
        template = schema.schema()
    else:
        template = build_schema_template(
            schema,
            pipeline_import_path=import_path,
            required_only=required_only,
            add_descriptions=add_descriptions,
        )

    if template_path:
        with open(template_path, "w") as file:
            json.dump(template, file, indent=2, default=jsonify)
    else:
        click.echo(json.dumps(template, indent=2, default=jsonify))


@click.command()
@click.argument("config_path", type=click.Path())
def validate_config(config_path: str) -> None:
    """Validate config without running a pipeline.

    \b
    Example:
        eogrow-validate config_files/config.json
    """
    for config in collect_configs_from_path(config_path):
        raw_config = interpret_config_from_dict(config)
        load_pipeline_class(config).Schema.parse_obj(raw_config)

    click.echo("Config validation succeeded!")


@click.command()
@click.argument("config_path", type=click.Path())
def run_test_pipeline(config_path: str) -> None:
    """Runs a test pipeline that only makes sure the managers work correctly. This can be used to select best
    area manager parameters.

    \b
    Example:
        eogrow-test any_pipeline_config.json
    """
    for crude_config in collect_configs_from_path(config_path):
        raw_config = interpret_config_from_dict(crude_config)
        pipeline = TestPipeline.with_defaults(raw_config)
        pipeline.run()


def _parse_cli_variable(mapping_str: str) -> Tuple[str, str]:
    """Checks that the input is of shape `name:value` and then splits it into a tuple"""
    match = re.match(r"(?P<name>.+?):(?P<value>.+)", mapping_str)
    if match is None:
        raise ValueError(f'CLI variable input {mapping_str} is not of form `"name:value"`')
    parsed = match.groupdict()
    return parsed["name"], parsed["value"]



"""
Utilities for solving different problems in `eo-grow` package structure, which are mostly a pure Python magic.
"""
from __future__ import annotations

import importlib
import re
from typing import TYPE_CHECKING, Any, Dict, Type

from packaging.requirements import Requirement

if TYPE_CHECKING:
    from ..core.pipeline import Pipeline
    from ..core.schemas import BaseSchema

_PIPELINE_PARAM_NAME = "pipeline"


def load_pipeline_class(config: dict) -> Type[Pipeline]:
    """Given a config object it loads the pipeline class referenced in the config"""
    pipeline_class_name = config.get(_PIPELINE_PARAM_NAME)
    if pipeline_class_name is None:
        raise ValueError(f"Config file is missing `{_PIPELINE_PARAM_NAME}` parameter, don't know which pipeline to use")

    pipeline_class = import_object(pipeline_class_name)
    return pipeline_class


def collect_schema(class_with_schema: Type) -> Type[BaseSchema]:
    """A utility that collects a schema from the given object.

    The object is expected to hold a unique internal class which inherits from `BaseSchema`. Example:

    class MyObject:
        class Schema(BaseSchema):
            ...

    This utility would provide `MySchema`. It works also if `MyObject` inherits from a class that holds the schema.
    """
    try:
        return class_with_schema.Schema
    except AttributeError as exception:
        raise SyntaxError(
            f"Class {class_with_schema} is missing a schema. Each `EOGrowObject` class needs to contain a pydantic "
            "model named `Schema`."
        ) from exception


def import_object(import_path: str) -> Any:
    """Imports an object from a given import path"""
    if "." not in import_path:
        raise ValueError(f"Import path `{import_path}` doesn't reference an object in a module.")
    module_name, object_name = import_path.rsplit(".", 1)

    try:
        module = importlib.import_module(module_name)
    except ModuleNotFoundError as exception:
        raise ModuleNotFoundError(f"{exception}. Given import path `{import_path}` is invalid.") from exception

    if hasattr(module, object_name):
        return getattr(module, object_name)

    raise ImportError(
        f"Cannot import name `{object_name}` from {module_name} ({module.__file__}). Given import path "
        f"`{import_path}` is invalid."
    )


def get_os_import_path(import_path: str) -> str:
    """For a Python import path it provides OS import path.

    E.g. `eogrow.utils.meta` -> `/home/ubuntu/.../eogrow/utils/meta.py`
    """
    module_spec = importlib.util.find_spec(import_path)
    if module_spec is not None and module_spec.origin is not None:
        return module_spec.origin
    raise ValueError(f"Given import path {import_path} not found")


def get_package_versions() -> Dict[str, str]:
    """A utility function that provides dependency package versions

    :return: A dictionary with versions
    """

    def is_base(req: str) -> bool:
        """Filters out packages needed only for development or documentation"""
        return re.search(r"extra == .(DEV|dev|DOCS|docs).", req) is None

    try:
        eogrow_reqs = importlib.metadata.requires("eo-grow") or []
        requirements = [Requirement(req).name for req in eogrow_reqs if is_base(req)]
        return {pkg: importlib.metadata.version(pkg) for pkg in ["eo-grow", *requirements]}

    except BaseException as ex:
        return {"error": repr(ex)}
